#! /usr/bin/env python

import os, sys, re, logging, collections

if 'USHhwrf' in os.environ:
    sys.path.append(os.environ['USHhwrf'])
elif 'HOMEhwrf' in os.environ:
    sys.path.append(os.path.join(os.environ['HOMEhwrf'],'ush'))
else:
    guess_HOMEhwrf=os.path.dirname(os.path.dirname(
            os.path.realpath(__file__)))
    guess_USHhwrf=os.path.join(guess_HOMEhwrf,'ush')
    sys.path.append(guess_USHhwrf)

import produtil.setup
import hwrf.launcher
from hwrf.numerics import to_datetime

logger=None
startdata="""
# Holdvars file with ksh variables:
holdvars="{holdvars}"

# Main conf file:
CONFhwrf="{CONFhwrf}"

# Cycle being run:
cycle={YMDH}

# Three character storm ID -- just number and basin letter:
stormid3="{vit[stormid3]}"

# Long storm ID:
longstormid="{vit[longstormid]}"
"""  # Don't forget the end of line before the """

def usage(logger):
    logger.critical('Invalid arguments to exhwrf_launch.py.  Aborting.')
    print '''
Usage: exhwrf_launch.py 2014062400 95E case_root /path/to/parm [options]

Mandatory arguments: 
  2014062400 -- the cycle to run
  95E -- the storm to run
  case_root -- FORECAST = real-time mode, HISTORY = retrospective mod
  /path/to/parm -- location of parm directory where standard conf files
      reside

Optional arguments:
section.option=value -- override conf options on the command line
/path/to/file.conf -- additional conf files to parse

Aborting due to incorrect arguments.'''
    sys.exit(2)

def main():
    produtil.setup.setup()
    logger=logging.getLogger('exhwrf_launch')
    PARAFLAG = ( 'YES' == os.environ.get('PARAFLAG','YES') )
    logger.info('Top of exhwrf_launch.')
    moreopt=collections.defaultdict(dict)

    args=sys.argv[1:]

    if len(args)<4: usage(logger)

    # Find cycle and stormid:
    cycle=to_datetime(args[0])
    stid=args[1].upper()
    if not re.match('^[0-9][0-9][ABCELPQSW]$',stid):
        logger.error('%s: invalid storm id.  Must be a three character '
                     'storm ID such as 90L or 13W'%(stid,))
        sys.exit(2)
    logger.info('Requested storm %s cycle %s'
                %(stid,cycle.strftime('%Y%m%d%H')))

    # Get the case root:
    case_root=args[2].upper()
    if case_root=='HISTORY':
        real_time=False
    elif case_root=='FORECAST':
        real_time=True
    else:
        logger.error('%s: invalid case root.  Must be HISTORY for '
                     'retrospective runs or FORECAST for real-time runs.'
                     %(case_root,))
        sys.exit(2)
    logger.info('Case root is '+repr(case_root))

    # Find the parm directory
    parm=args[3]
    if not os.path.exists(parm):
        logger.error(parm+': parm directory does not exist')
        sys.exit(2)
    elif not os.path.isdir(parm):
        logger.error(parm+': parm directory is not a directory')
        sys.exit(2)
    infiles=[ os.path.join(parm,'hwrf_input.conf'),
              os.path.join(parm,'hwrf.conf'), 
              os.path.join(parm,'hwrf_holdvars.conf'), 
              os.path.join(parm,'hwrf_basic.conf'),
              os.path.join(parm,'system.conf')
              ]
    
    # Now look for any option and conf file arguments:
    logger.info('Scan %d optional arguments.'%(len(args)-3))
    args=args[4:]
    bad=False
    for iarg in xrange(len(args)):
        logger.info(args[iarg])
        m=re.match('''(?x)
          (?P<section>[a-zA-Z][a-zA-Z0-9_]*)
           \.(?P<option>[^=]+)
           =(?P<value>.*)$''',args[iarg])
        if m:
            logger.info('Set [%s] %s = %s'%(
                    m.group('section'),m.group('option'),
                    repr(m.group('value'))))
            moreopt[m.group('section')][m.group('option')]=m.group('value')
        elif os.path.exists(args[iarg]):
            logger.info('%s: read this conf file'%(args[iarg],))
            infiles.append(args[iarg])
        else:
            bad=True
            logger.error('%s: invalid argument.  Not an config option '
                         '(a.b=c) nor a conf file.'%(args[iarg],))
    if bad:
        sys.exit(2)

    for file in infiles:
        if not os.path.exists(file):
            logger.error(file+': conf file does not exist.')
            sys.exit(2)
        elif not os.path.isfile(file):
            logger.error(file+': conf file is not a regular file.')
            sys.exit(2)
        elif not produtil.fileop.isnonempty(file):
            logger.warning(
                    file+': conf file is empty.  Will continue anyway.')
        logger.info('Conf input: '+repr(file))

    conf=hwrf.launcher.launch(infiles,cycle,stid,moreopt)
    conf.sanity_check()

    holdvars=conf.strinterp('dir','{com}/{stormlabel}.holdvars.txt')
    logger.info(holdvars+': write holdvars here')
    with open(holdvars,'wt') as f:
        f.write(conf.make_holdvars())
        if PARAFLAG and conf.has_section('kick_scripts'):
            write_kick_holdvars(f,conf)

    if conf.has_option('config','startfile'):
        startfile=conf.getstr('config','startfile')
        logger.info(startfile+': Write holdvars and conf location here.')
        startcontents=conf.strinterp('config',startdata,holdvars=holdvars)
        with open(startfile,'wt') as f:
            f.write(startcontents)

def write_kick_holdvars(f ,conf):
    keepme=set(['ACCOUNTLL','GROUPLL','CLASSLL','JET_NAME'])
    for var,val in conf.items('kick_scripts'):
        val=str(val)
        if var[-7:] == 'CLASSLL' or var in keepme:
            f.write('export %s="%s"\n'%(var,val))

if __name__ == '__main__': main()
