#! /bin/env python

"""This script implements data delivery functionality specific to the
NCEP HWRF system.  It is run after everything except archiving, and it
copies the output of some prior jobs to the COM directory under
different names.  In the operational HWRF, this job is also supposed
to email the NOAA Senior Duty Meterologist the HWRF AFOS file.
However, that functionality is intentionally left out at the moment.

This job is written specifically for the 2014 and 2013 operational
HWRF systems.  It does recognize the run_gsi, run_ocean and
run_relocation options, but it will need to be modified to handle
other configurations.  However, this job is optional.  If you do not
run this job, the later archiving and graphics jobs will still run
correctly."""

import os, glob, sys, logging
import produtil.fileop, produtil.setup

import produtil.log, produtil.setup, produtil.datastore
from produtil.log import jlogger

def deliver_multi(sourceglob,target,logger):
    count=0
    if sourceglob.find('*')>=0 or sourceglob.find('?')>=0:
        logger.info('Globbing for: %s'%(sourceglob,))
        for source in glob.glob(sourceglob):
            produtil.fileop.deliver_file(source,target,keep=True,
                                         logger=logger)
            count+=1
    else:
        produtil.fileop.deliver_file(sourceglob,target,keep=True,
                                     logger=logger)
        count=1
    return count

class Deliverer(object):
    def __init__(self,logger,conf):
        """Creates a new Deliverer with the specified logging.Logger
        and HWRFConfig."""
        self.__logger=logger
        self.__conf=conf
        self.__failures=0 # Count of the number of failed deliveries
        self.__morevars=dict() # dict of string format substitutions
    def log(self,sublog=None):
        """Gets the logging.Logger for this Deliverer.  If "sublog" is
        provided, a new logging.Logger is created for that subdomain
        of this Deliverer's logging domain."""
        if sublog is not None:
            return logging.getLogger(self.__logger.name+'.'+sublog)
        return self.__logger
    def __setitem__(self,key,value):
        """Sets a key in an internal dict of values used for string
        formatting."""
        self.__morevars[key]=value
    def __getitem__(self,key):
        """Gets a key from the hash of internal values used for string
        formatting."""
        return self.__morevars[key]
    @property
    def failures(self):
        """The number of failed deliveries."""
        return self.__failures
    def reset(self):
        """Resets the number of failures to zero."""
        self.__failures=0
    @property
    def conf(self):
        """The HWRFConfig object for this Deliverer."""
        return self.__conf
    def deliver_file(self,workfile,comfile=None,from_com=False,
                     optional=False,**kwargs):
        """Delivers the specified work area file to the specified com
        location.  OR, if from_com=True, it does the opposite: deliver
        from com to the workfile.  If the com location is not
        specified, a suitable default will be chosen.  If
        optional=True, deliver_file will ignore the file if it is
        missing.  Otherwise, a missing file counts as a failure in the
        internal failure counter.  Additional keyword arguments are
        sent to conf.strinterp.  If the workfile is a relative path,
        it is relative to the WORKhwrf directory.

        When delivering to COM (from_com=False, the default), the
        workfile can instead be a Product, in which case the
        Product.location and .available are checked for availability
        information.  When delivering from COM (from_com=True), the
        workfile must be a string path."""
        conf=self.conf
        logger=self.log()
        if kwargs:
            morevars=dict(self.__morevars)
            morevars.update(kwargs)
        else:
            morevars=self.__morevars
        if not from_com and isinstance(workfile,produtil.datastore.Product):
            with workfile.dstore.transaction() as t:
                loc=workfile.location
                av=workfile.available
                if loc is None or loc=='':
                    logger.error('%s: no location'%(workfile.did,))
                    self.__failures+=1
                    return False
                elif not av:
                    logger.error('%s: not available (location=%s)'%(
                            workfile.did,workfile.location))
                    self.__failures+=1
                    return False
                else:
                    workpath=workfile.location
        elif from_com:
            assert(isinstance(workfile,basestring))
            workpath=conf.strinterp('config',workfile,**morevars)
            logger.info('from_com workpath='+workpath)
        else:
            assert(isinstance(workfile,basestring))
            workpath=os.path.join(
                conf.getdir('WORKhwrf'),
                conf.strinterp('config',workfile,**morevars) )
            
        if comfile is None:
            if from_com:
                raise AssertionError('When copying from com, the comfile '
                                     'must be specified.')
            compath=os.path.join(
                conf.getdir('com'),
                conf.strinterp('config','{out_prefix}.{workbasename}',
                               workbasename=os.path.basename(workpath)))
        else:
            compath=os.path.join(
                conf.getdir('com'),
                conf.strinterp('config',comfile,**morevars) )
        logger.info('compath is '+compath)
        if from_com and optional and not os.path.exists(compath):
            logger.info('Optional file does not exist: '+repr(compath))
            return
        elif not from_com and optional and not os.path.exists(workpath):
            logger.info('Optional file does not exist: '+repr(workpath))
            return
        try:
            if from_com:
                frompath=compath
                topath=workpath
            else:
                frompath=workpath
                topath=compath
            logger.info('deliver %s => %s'%(frompath,topath))
            count=deliver_multi(frompath,topath,logger=logger)
        except EnvironmentError as e:
            logger.error('%s: cannot deliver: %s\n'%(workpath,str(e)),
                         exc_info=True)
            self.__failures+=1
            return False
        if count==0:
            if optional:
                logger.warning('%s: no files matched this glob.  This '
                               'file was optional, so I will continue.'
                               %(frompath,))
            else:
                logger.error('%s: no files matched this glob.'%(frompath,))
                self.__failures+=1
                return False
        return True

########################################################################

def main():
    import hwrf_expt
    hwrf_expt.init_module()

    conf=hwrf_expt.conf
    relocation=conf.getbool('config','run_relocation',True)
    coupled=conf.getbool('config','run_ocean',True)
    GSI=conf.getbool('config','run_gsi') 

    logger=conf.log('output')

    hwrf_expt.wrfcopier.run()

    D=Deliverer(logger,conf)
    D['wrfdir']=hwrf_expt.runwrf.workdir

    if GSI:
        D['gsi_d02']=hwrf_expt.gsi_d02.outdir
        D['gsi_d03']=hwrf_expt.gsi_d03.outdir

    logger.info('WRF run directory is %s'%(repr(D['wrfdir']),))

    D.deliver_file('{WORKhwrf}/jlogfile',optional=True)

    d01=hwrf_expt.moad
    d02=hwrf_expt.storm1outer
    d03=hwrf_expt.storm1inner

    if GSI:
        # Copy the original wrfinput file before DA:
        org_d01=hwrf_expt.gfs_init.realinit.wrfinput_at_time(
            hwrf_expt.cycle,d01)
        D.deliver_file(org_d01,'{out_prefix}.wrfinput_d01_org')

    if GSI:
        # Copy the original wrfanl files before DA:
        rinit=hwrf_expt.fgat_init.init_at_time(hwrf_expt.conf.cycle)
        org_d02=rinit.rstage3.wrfanl_at_time(hwrf_expt.conf.cycle,d02)
        org_d03=rinit.rstage3.wrfanl_at_time(hwrf_expt.conf.cycle,d03)
        D.deliver_file(org_d02,'{out_prefix}.wrfanl_d02_org')
        D.deliver_file(org_d03,'{out_prefix}.wrfanl_d03_org')

    for domain in hwrf_expt.gfs_init.runwrfanl.sim:
        if not domain.is_moad():
            org_prod=hwrf_expt.gfs_init.runwrfanl.wrfanl_at_time(
                hwrf_expt.cycle,domain)
            D.deliver_file(org_prod,'{out_prefix}.wrfanl_d{gid:02d}_org',
                         gid=int(domain.get_grid_id()))

    if GSI:
        D.deliver_file('{gsi_d02}/satbias_out',
                       '{out_prefix}.gsi_cvs2.biascr')
        D.deliver_file('{gsi_d03}/satbias_out',
                       '{out_prefix}.gsi_cvs3.biascr')

    if coupled:
        D.deliver_file('{wrfdir}/MDstatus',optional=True)
        for ocrest in ( '0000.nc', '0001.nc', '0002.nc', '0003.nc', 
                        '0005.nc',
                        'el_initial.nc', 'grid.nc', 'ts_clim.nc', 
                        'ts_initial.nc', 'uv_initial.nc' ):
            D.deliver_file('{wrfdir}/{vit[stormname]}.{ocrest}',
                           '{out_prefix}.pom.{ocrest}', ocrest=ocrest)

    logcount=0
    for ext in ('log','out','err'):
        globme=conf.getdir('WORKhwrf')+'/*.'+ext
        logger.info('Globbing for %s log files'%(globme,))
        for log in glob.glob(globme):
            logcount+=1
            D.deliver_file(log)
    logger.info('Found %d log file(s)'%(logcount,))

    if D.failures>0:
        jlogger.critical('HWRF: unable to deliver some products to com.')
        sys.exit(1)

    D.reset()
    def fromcom(workpath,compath,optional=False):
        D.deliver_file(workpath,compath,from_com=True,optional=optional)

    def havedir(sdir):
        there=conf.get('dir',sdir,'NOPE')
        if there=='NOPE':
            return False
        produtil.fileop.makedirs(there)
        return True

    if havedir('outatcf'):
        fromcom('{outatcf}','{out_prefix}.trak.hwrf.atcfunix')
    if havedir('outdiag'):
        fromcom('{outdiag}','{out_prefix}.trak.3hourly*')
        fromcom('{outdiag}','{out_prefix}*resolution',True)
        fromcom('{outdiag}','{out_prefix}*htcf*stats',True)
        fromcom('{outdiag}','{out_prefix}*htcf',True)
        fromcom('{outdiag}','a*.dat')
        fromcom('{outdiag}','{out_prefix}.stats.tpc',optional=True)
    if havedir('outships'):
        fromcom('{outships}','figures/*.txt',optional=True)
    if havedir('outstatus'):
        fromcom('{outstatus}','{WORKhwrf}/submit.out')
        timings=conf.strinterp('config','{outstatus}/{out_prefix}.timings')
        inout=conf.strinterp('config','{WORKhwrf}/hwrf_*.out')
        with open(timings,'wt') as outf:
            for inoutfile in glob.glob(inout):
                if not os.path.exists(inoutfile):
                    logger.warning('%s: file does not exist; skipping'
                                   %(inoutfile,))
                with open(inoutfile,'rt') as inf:
                    for line in inf:
                        if line.find('TIMING')>=0:
                            print>>outf,line.rstrip()
        fromcom('{outstatus}','{WORKhwrf}/submit.out')
    if havedir('outatcfcorrected'):
        inatcf=conf.strinterp('config',
                              '{com}/{out_prefix}.trak.hwrf.atcfunix')
        outatcf=conf.strinterp(
            'config','{outatcfcorrected}/{out_prefix}.trak.hwrf.atcfunix')
        with open(inatcf,'rt') as inf:
            with open(outatcf,'wt') as outf:
                for line in inf:
                    line=line.rstrip() \
                        .replace('AA','IO')\
                        .replace('BB','IO')\
                        .replace('SP','SH')\
                        .replace('SI','SH')
                    print>>outf,line

########################################################################

if __name__=='__main__':
    try:
        produtil.setup.setup()
        jlogger.info('hwrf_output is starting')
        main()
        jlogger.info('hwrf_output has completed')
    except Exception as e:
        jlogger.critical('hwrf_output is aborting: '+str(e),exc_info=True)
        sys.exit(2)
