"""This module contains exceptions raised by the hwrf package.  The
only exceptions are the hwrf.revital and hwrf.storminfo modules which
contain their own exceptions for backward compatibility to old
scripts."""

class HWRFError(Exception): 
    """This is the base class of exceptions raised by the HWRF module
    due to HWRF-specific failures.  It is possible to get other
    exceptions from the StandardException heirarchy in certain failure
    cases.  For example, trying to obtain the fifth of three domains
    may raise KeyError.  Also, see pom.exceptions for exceptions that
    may be raised from the pom package."""

class DuplicateTaskName(HWRFError):
    """Raised when more than one task is registerred with the same
    name in an HWRFConfig object."""

########################################################################
# ARCHIVING EXCEPTIONS


########################################################################
# NAMELIST-RELATED EXCEPTIONS

class NamelistValueError(ValueError): 
    """Raised when hwrf.namelist cannot convert a value to or from
    Fortran namelist format."""
class NamelistKeyError(KeyError):
    """Raised when an hwrf.namelist is asked for a key that does not
    exist."""
    def __init__(self,message,section,var):
        super(NamelistKeyError,self).__init__(message)
        self.section=section
        self.var=var
        self.message=message
    def __str__(self):
        if self.section=='-trait-':
            return 'trait %s: %s' % (self.var,self.message)
        else:
            return '&%s %s: %s' % (self.section,self.var,self.message)
    def __repr__(self):
        return 'NamelistError(%s,%s,%s)' % \
            ( repr(self.message), repr(self.section), repr(self.var) )

########################################################################
# SANITY CHECKER EXCEPTIONS

class HWRFSanityError(HWRFError):
    """Base class of all sanity checker exceptions."""
class HWRFDirInsane(HWRFSanityError):
    """Raised when a directory is unspecified, missing or invalid."""
    def __init__(self,message,dir):
        super(HWRFDirInsane,self).__init__(message)
        self.dir=dir
class HWRFConfigInsane(HWRFSanityError):
    """Raised when the requested configuration fails a sanity check."""
class HWRFConfigUnsupported(HWRFConfigInsane):
    """Raised when the user requests a configuration that makes sense,
    but is not supported."""
class HWRFStormInsane(HWRFSanityError):
    """Raised when the configuration had a different storm than expected."""
class HWRFCycleInsane(HWRFSanityError):
    """Raised when the configuration had a different cycle than expected."""
class HWRFVariableInsane(HWRFSanityError):
    """Raised when a sanity check on a variable's value failed."""
class HWRFInputInsane(HWRFSanityError):
    """Raised when input files to HWRF fail a sanity check."""

########################################################################
# OCEAN EXCEPTIONS

#NOTE: See pom.exceptions for more
class OceanInitFailed(HWRFError):
    """Raised when the ocean init did not produce some expected outputs."""

########################################################################
# GSI EXCEPTIONS

class GSIInputError(HWRFError):
    """Raised when GSI cannot find a required input file."""
    
########################################################################
# TRACKER EXCEPTIONS

class TrackerError(HWRFError): 
    """Base class of hwrf.tracker exceptions."""
class TrackerModeError(TrackerError): 
    """Raised when an impossible tracker configuration is requested,
    such as running with a grid that is both regional and global."""
class TrackerStormError(TrackerError): 
    """Raised when multiple storms are requested, but only one was
    expected."""
class TrackerInputError(TrackerError): 
    """Base class of exceptions raised when the tracker's input files
    are missing or invalid."""
class MissingGRIBError(TrackerInputError): 
    """Not currently used, this would be raised when GRIB inputs to
    the tracker are missing."""
class GRIBLocationError(TrackerInputError): 
    """Raised when no location is specified for a tracker input GRIB
    file."""

########################################################################
# TIME-RELATED EXCEPTIONS (used by many modules)

class HWRFTimeError(HWRFError): 
    """Base class used for time-related HWRF exceptions."""

# Time and timestep exceptions:
class InvalidTimestep(HWRFTimeError): 
    """Raised when a timestep is invalid, such as a negative timestep
    for a situation that requires a positive one."""
class OverspecifiedOutputTime(HWRFTimeError): 
    """Raised when an output time is specified in two redundant ways.
    For example, one could specify a forecast time directly, and also
    specify the analysis time and forecast time delta."""
class NoOutputTime(HWRFTimeError): 
    """Raised when a time was required, but none was provided."""
class TimezoneProvided(HWRFTimeError): 
    """Raised when a timezone is provided.  The hwrf package does not
    support timezones: all times are in UTC."""
class PrecisionTooHigh(HWRFTimeError): 
    """Raised when a time was provided that contained fractions of a
    second, for a function that cannot handle that.  For example, the
    WRF output files must be exactly on a second boundary."""
class NotInTimespan(HWRFTimeError): 
    """Raised when a time is outside the range of times being
    processed by a function."""
class NoNearbyValues(HWRFTimeError): 
    """Raised when an operation has a set of known times, but another
    provided time is not near one of those known times."""

class InvalidTimespan(HWRFTimeError):
    """Superclass of exceptions relating to groups of one or more
    distinct times and relationships between them."""
    def __init__(self,message,start,end):
        super(InvalidTimespan,self).__init__(message)
        self.start=start
        self.end=end
class EndBeforeStart(InvalidTimespan): 
    """Raised when the end of a timespan is before the beginning."""
class EndNotTimestep(InvalidTimespan): 
    """Presently unused, this was to indicate that the end of a
    timespan is not on a timestep, for temporally discrete
    processes."""
class StartNotAtParentTimestep(InvalidTimespan): 
    """Raised when a timespan's beginning is not at a timestep."""
class TimestepTooLong(InvalidTimespan): 
    """Raised when a timestep is too long for the process under
    consideration."""
class TimestepTooShort(InvalidTimespan): 
    """Raised when a timestep is too short for the process under
    consideration."""
class NoTimespan(InvalidTimespan): 
    """Raised when a timespan was expected, but none was available."""

########################################################################
# REGRIB-RELATED EXCEPTIONS (mostly hwrf.regrib and hwrf.gribtask)

class RegribError(HWRFError): 
    """Superclass of errors used by Regrib."""
class GRIBInputError(RegribError): 
    """Raised when a GRIB file is invalid.  This can be due to either
    an input to an operation, or the output from the operation."""
class Subsetless(RegribError): 
    """Raised when a Regrib was expecting a GRIB subsetter function,
    but no such function was provided."""

class InvalidRegribResult(RegribError): 
    """Part of debug assertions in hwrf.regrib, this is raised when
    the wrong type is generated by the "make" function."""

class RegribProductError(RegribError): 
    """Superclass of errors relating to regrib products."""
class NoProductError(RegribProductError): 
    """Raised when an operation that produces input to Regrib should
    have produced a Product, but produced nothing at all."""
class ProductAmbiguityError(RegribProductError): 
    """Raised when an operation that provides input to Regrib produces
    more than one product."""
class NoIndexError(RegribError):
    """Raised when a GRIB file should have an index file already, but
    doesn't."""

class RegribManyError(RegribError): 
    """Base class of errors from the hwrf.regrib.RegribMany"""
class RegribKeyError(RegribManyError): 
    """Raised when a RegribMany is given an invalid name: one that
    does not match a known grid or operation."""

class RegribGridError(RegribError): 
    """Base class of grid-related regrib errors."""
class GridlessError(RegribGridError): 
    """Raised when a grid was expected but none was provided."""
class GridMismatchError(RegribGridError): 
    """Raised when two GRIB files have non-matching grids, but a
    matching grid is required."""

class GribberError(RegribError):
    """Raised by GRIBTask for unexpected errors that did not come from
    the underlying RegribAll object.  This is used in GRIBTask.run's
    "run it now" mode, when setting raiseall=True."""

########################################################################
# WRF EXCEPTIONS

class WRFError(HWRFError): 
    """Base class of WRF-related errors."""
class DomainsDone(WRFError): 
    """Raised when it is no longer possible to add domains, but the
    caller tried to add one."""
class WRFDomainError(WRFError):
    """Base class of errors related to a WRF domain."""
    def __init__(self,message,domainname=None):
        super(WRFDomainError,self).__init__(message)
        self.domainname=domainname
class InvalidDomainStart(WRFDomainError): 
    """Raised when the conf value specifying the type of domain parent
    start information contains invalid data (not "fixed" or
    "auto.")"""
class DomainExists(WRFDomainError): 
    """Raised when adding a domain, if a domain by the same name
    already exists."""
class OutputStreamDisabled(WRFError): 
    """Raised when attempting to obtain information about when a WRF
    stream outputs, if the stream is disabled."""
class RealNMMError(WRFError): 
    """Raised when failing to run the real_nmm program."""
class SetNestFailed(WRFError): 
    """Raised when failing to set the parent start location via the
    set_nest (set_ij_start) program."""

########################################################################
# FORECAST INPUT DATA EXCEPTIONS

class ForecastInputError(HWRFError): pass
class WRFMissing(ForecastInputError):
    """Raised when wrf or real want to run but needed input data is missing."""
class WRFGeogridMissing(WRFMissing):
    """Raised when wrf or real want to run but geogrid data is missing"""
class WRFMetgridMissing(WRFMissing):
    """Raised when wrf or real want to run, but metgrid data is missing."""
class WRFInputMissing(WRFMissing):
    """Raised when wrf or real want to run, but wrfinput files are missing."""
class WRFBdyMissing(WRFMissing):
    """Raised when wrf or real want to run, but wrfbdy files are missing."""
class WRFAnlMissing(WRFMissing):
    """Raised when wrf or real want to run, but wrfanl files are missing."""
class WRFPrepMissing(WRFMissing):
    """Raised when real wants prep_hybrid data, but that data is missing"""

########################################################################
# WPS EXCEPTIONS

class WPSError(HWRFError): 
    """Base class of WPS-related exceptions."""
class WPSAssertion(HWRFError):
    """Raised when WPS inputs are implausable."""
    def __init__(self,message,**kwargs):
        HWRFError(self,message,**kwargs)
class GeogridNoLog(WPSError):
    """Raised when geogrid did not create a log file."""
class GeogridNoOutput(WPSError):
    """Raised when a geogrid output file is missing or empty."""
class UngribNoInput(WPSError):
    """Raised when ungrib cannot find an input file it needs."""
class UngribSubsetError(WPSError):
    """Raised when wgrib or wgrib2 generates an empty or invalid file."""

########################################################################
# PrepHybrid exceptions:

class PrepHybridError(HWRFError): 
    """Base class of exceptions related to the prep_hybrid program."""
class NoGeogData(PrepHybridError): 
    """Raised when the prep_hybrid program cannot find WPS geogrid
    output data."""

########################################################################
# INPUT EXCEPTIONS

class HWRFInputError(HWRFError): 
    """Base class of exceptions related to the hwrf.input module."""
class InputSourceBadType(HWRFInputError):
    """Raised when a configuration file requests a DataCatalog class
    that does not exist."""
class PartialTransfer(HWRFInputError):
    """Raised when a file transfer, done by an InputSource, was
    incomplete."""
class UnsupportedTransfer(HWRFInputError):
    """Raised when a configuration file requests an unsupported data
    transfer method (such as carrier pigeon)."""

########################################################################
# Post exceptions

class PostFailed(HWRFError):
    """Raised upon errors that would cause a retry, in the
    PostOneWRF.run when passed the raiseall=True argument."""

class PostHasNoInput(HWRFError):
    """Raised when the post's input file is not available and
    raiseall=True in PostOneWRF.run"""

########################################################################
# Relocation exceptions

class RelocationError(HWRFError):
    """Raised when something in the vortex relocation fails."""
class RelocateOutputMissing(RelocationError):
    """Raised when a relocation program did not produce an expected
    output file."""
class StormRadiusError(RelocationError):
    """Raised when the merge cannot find the storm_radius file in the
    relocate or fix directory.."""
class NoSuchDomain(RelocationError):
    """Raised by hwrf.input when trying to get the wrong domain from
    its hwrf.relocate.Relocate child objects."""
