import os, shutil
import produtil.fileop, produtil.datastore, produtil.cd, produtil.run
import hwrf.hwrftask, hwrf.wrf, hwrf.numerics, hwrf.namelist
import datetime
import pdb;

from produtil.datastore import COMPLETED
from produtil.cd import NamedDir
from produtil.fileop import make_symlink, deliver_file, isnonempty, make_symlinks_in
from produtil.run import mpirun, mpi, openmp, checkrun
from hwrf.numerics import to_datetime,to_datetime_rel, to_fraction, to_timedelta
from hwrf.exceptions import GSIInputError

class GSIBase(hwrf.hwrftask.HWRFTask):
    """Base class of anything that runs the GSI.  Do not use directly."""
    def __init__(self,dstore,conf,section,domain,wrf_in_prod,sim,
                 taskname=None,atime=None,parent_atime=None,**kwargs):
        """The GSIBase constructor:
           dstore -- passed to Datum: the Datastore object for this Task
           conf -- the conf object for this task (passed to HWRFTask)
           section -- the conf section for this task (passed to HWRFTask)
           domain -- the WRFDomain for this GSI.  Must have been
             initialized by a WRFSimulation
           wrf_in_prod -- the Product for the wrfinput_d01 or
             ghost_d0* file for that domain
           taskname -- Optional: the taskname for this product.
             Passed to HWRFTask
           atime -- the analysis time as a datetime.datetime.
             Default: conf.cycle
           kwargs -- ignored; passed to HWRFTask"""
        super(GSIBase,self).__init__(dstore,conf,section,taskname=taskname,
                                     **kwargs)
        taskname=self.taskname
        if atime is None: atime=conf.cycle
        if parent_atime is None:
            parent_atime=to_datetime_rel(-6*3600,atime)
        self._atime=to_datetime(atime)
        self._parent_atime=to_datetime(parent_atime)
        self._domain=domain
        self._wrf_in_prod=wrf_in_prod
        self._sim=sim
        assert(wrf_in_prod is not None)
        assert(wrf_in_prod.location is not None)
        assert(wrf_in_prod.location!='')

        # Generate the FileProduct object for our output file:
        with dstore.transaction() as t:
            prodname='gsi_out_'+self.domain.name
            prod=produtil.datastore.FileProduct(dstore,prodname,taskname,
                 location=os.path.join(self.outdir,prodname))
            self._wrf_out_prod=prod

        # Get the datasets and items for our data grabbers:
        def getc(what,default):
            if what in kwargs:
                return str(kwargs[what])
            s=self.confstr(what,'')
            if s is not None and s!='':
                return s
            return default
        #self._sigma_dataset=getc('sigma_dataset','gfs')
        #self._sigma_item=getc('sigma_item','gfs_sf')
        self._sigma_dataset=getc('sigma_dataset','gdas1')
        self._sigma_item=getc('sigma_item','gdas1_sf')
        self._bufr_dataset=getc('bufr_dataset','gfs')
        self._bufr_item=getc('bufr_item',self._bufr_dataset+'_bufr')
        self._prepbufr_item=getc('prepbufr_item','prepbufr_nr')
        self._enkf_dataset=getc('enkf_dataset','enkf')
        self._enkf_item=getc('enkf_item','enkf_sfg')
        self._biascr_dataset=getc('biascr_dataset','gdas1')
        self._biascr_item=getc('biascr_item','gdas1_biascr')
        self._biascr_pc_item=getc('biascr_pc_item','gdas1_biascr')
        self._abias_item=getc('abias_item','gdas1_abias')
        self._satang_item=getc('satang_item','gdas1_satang')
        self._biascr_pc_item=getc('biascr_item','gdas1_biascr')

        # Get the DataCatalog for our data grabbers:
        self._in_catalog=None
        incat_name=None
        if 'in_catalog' in kwargs:
            ink=kwargs['in_catalog']
            if isinstance(ink,hwrf.input.DataCatalog):
                self._in_catalog=ink
            elif isinstance(ink,basestring):
                incat_name=ink
            elif ink is None: pass
            else:
                raise TypeError('In hwrf.gsi.GSIBase.__init__, in_catalog must be None, a basestring or a DataCatalog.  You provided an object of type %s, with value %s.'%(type(ink).__name__,repr(ink)))
        if self._in_catalog is None:
            if incat_name is None:
                incat_name=self.confstr('catalog')
            self._in_catalog=hwrf.input.DataCatalog(self.conf,incat_name,self.parent_atime)

    @property
    def wrf_top_Pa(self):
        """The WRF model top in pascals"""
        p_top_requested=self._sim.nl.nl_get('domains','p_top_requested')
        return float(p_top_requested)

    @property
    def atime(self): 
        """The analysis time of this GSI and of its parent global model."""
        return self._atime

    @property
    def parent_atime(self): 
        """The analysis time of the parent data assimilation models
        (ie.: GDAS, GFS ENKF). This is usually six hours before
        self.atime."""
        return self._parent_atime

    def grab_bufr(self,atime=None,morevars=None):
        """Links or copies all needed bufr files to the local
        directory.  If sat_da is False, satellite obs will be
        omitted."""
        olist=self.confstr('obstypes')
        touched=set()
        for osection in olist.split(','):
            trim=osection.strip()
            if len(trim)>0 and not trim in touched:
                self.grab_obstype_section(trim,atime=atime,morevars=morevars)

    def grab_obstype_section(self,section,atime=None,morevars=None):
        """Copies or links observations specified in the obstype
        sections to the current working directory.  

        The section listed in self.section should contain an obstype
        option, whose value is a comma separated list of section
        names.  This method reads every section in that list.  For
        each section, the option keys are the local directory
        filenames expected by GSI, while the values are the data type
        part of the operational filename (ie.: the satwind in
        gfs.t12z.tm00.satwind.bufr_d).  There are a few special keys:

          dataset - the name of the dataset for hwrf.input purposes
          item - the name of the item for hwrf.input purposes
          type - the type of observation: satellite, or anything else.
            At present, only "satellite" has any special meaning.

        If the type is "satellite" then the entire section will be
        skipped if sat_da=False in this task's config section.

        Once the section is parsed, the files are all linked to this
        directory."""
        logger=self.log()
        if not isinstance(section,basestring): section=str(section)

        if atime is None: 
            atime=self.atime
        else:
            atime=to_datetime_rel(atime,self.atime)

        dataset=self.conf.get(section,'dataset')
        item=self.conf.get(section,'item')
        otype=self.conf.get(section,'type').lower()

        logger.warning('process obs section %s with dataset=%s item=%s '
                       'type=%s'%(section,dataset,item,otype))

        if otype=='satellite':
            if self.confbool('sat_da',True):
                logger.warning('%s: satellite data and satellite DA '
                               'is enabled.  Continuing...'%(section,))
            else:
                logger.warning('%s: satellite DA is disabled in hwrf.conf.  '
                               'Not linking bufr files from this section.'
                               %(section,))
                return


        obstypes=list()
        items=self.conf.items(section)
        otdict=dict( [ (v,k) for k,v in items ] )
        namer=lambda f,t: otdict[t]

        for localname,obstype in items:
            if localname in ['dataset','item','type']: continue
            obstypes.append(obstype)

        for obstype in obstypes:
            logger.warning('Find obstype=%s in dataset=%s item=%s'
                           %(obstype,dataset,item))
            if not isinstance(obstype,basestring):
                raise TypeError(
                    'In gsi.GSIBase.link_bufr, the obstypes parameter must '
                    'be an iterable container of basestrings.  One of the '
                    'elements was a %s (value %s) instead.'
                    %(type(obstype).__name__,repr(obstype)))
            there=self._in_catalog.locate(dataset,item,atime=atime,
                                          logger=logger,obstype=obstype)
            if there is None or there=='':
                msg='%s: Could not find a location for this obstype.'\
                    %(obstype,)
                logger.warning(msg)
                if self.confbool('require_all_bufrs',False):
                    raise GSIInputError(msg)
            elif produtil.fileop.isnonempty(there):
                bn=os.path.basename(there)
                on=namer(bn,obstype)
                make_symlink(there,on,logger=logger,force=True)
            else:
                msg='%s: Observation file is empty or non-existant: %s'\
                    %(obstype,there)
                logger.warning(msg)
                if self.confbool('require_all_bufrs',False):
                    raise GSIInputError(msg)
            

    def grab_prepbufr(self,atime=None,**kwargs):
        """Links or copies the prepbufr file to the local directory."""
        if atime is None:
            atime=self.atime
        else:
            atime=to_datetime_rel(atime,self.atime)
        logger=self.log()
        there=self._in_catalog.locate(self._bufr_dataset,self._prepbufr_item,
            atime=atime,logger=logger,**kwargs)
        if there is None or there=='':
            msg='Could not find the prepbufr file (item=%s dataset=%s)' \
                %(repr(self.prepbufr_item),repr(self._bufr_dataset))
            logger.warning(msg)
            raise GSIInputError(msg)
        elif not produtil.fileop.isnonempty(there):
            msg=there+': is non-existant or empty'
            logger.error(msg)
            raise GSIInputError(msg)
        make_symlink(there,'prepbufr',logger=logger,force=True)

    def grab_gfs_enkf(self,atime=None,**kwargs):
        """Links the GFS ENKF files to this directory"""
        logger=self.log()
        maxmemb=self.confint('num_enkf',80) 
        if maxmemb==0: 
            logger.warning('Number of ENKF members requested is 0.  Will '
                           'not link anything.')
            return
        assert(maxmemb>=0) # FIXME: REMOVE

        fhour=self.conffloat('enkf_fhr',6.0)
        enkf_age=self.conffloat('enkf_age_hr',6.0)
        my_atime=self.atime
        assert(my_atime is not None)
        atime=to_datetime_rel(-enkf_age*3600.0,my_atime)
        ftime=to_datetime_rel(3600*fhour,atime)
        with open('filelist','wt') as fl:
            for zmemb in xrange(maxmemb):
                imemb=1+zmemb
                there=self._in_catalog.locate(self._enkf_dataset,
                    self._enkf_item,atime=atime,logger=logger,
                    enkfmem=imemb,ftime=ftime,**kwargs)
                local=self.conftimestrinterp(
                    'sfg_{aYMDH}_fhr{fahr:02d}s_mem{imemb:03d}',
                    atime=atime,ftime=ftime,imemb=imemb)
                if not produtil.fileop.isnonempty(there):
                    logger.error(there+': is empty or non-existant')
                produtil.fileop.make_symlink(there,local,force=True,
                                             logger=logger)
                fl.write(local+'\n')

    def copy_wrf_inout(self,filename='wrf_inout'):
        """Copies the WRF analysis or input file to the specified filename"""
        logger=self.log()
        def namer(p,logger,*args): return filename
        def actor(p,name,logger,*args): 
            deliver_file(p.location,name,logger=logger)
        produtil.datastore.wait_for_products([self._wrf_in_prod],
            logger,namer,actor)

    def get_ghost(self,domain):
        """If this GSI is being run on the specified domain, returns
        the output product of GSI, otherwise returns None."""
        if domain==self.domain:
            return self._wrf_out_prod
        return None

    def get_wrfanl(self,domain):
        """If this GSI is being run on the specified domain, returns
        the output product of GSI, otherwise returns None."""
        if domain==self.domain:
            return self._wrf_out_prod
        return None

    def get_wrfinput(self):
        """If this GSI is being run on the WRF outermost domain
        (MOAD), returns the output product of GSI.  Otherwise, returns
        None."""
        if self.domain.is_moad():
            return self._wrf_out_prod
        else:
            return None
    @property
    def domain(self): 
        """The WRF domain for which GSI is being run."""
        return self._domain

    def products(self,domains=None,prodtype='wrf_out_prod'):
        """Iterates over all output products of this Task."""
        if domains is None:
            yield self._wrf_out_prod
        else:
            for domain in domains:
                if domain==self.domain:
                    yield self._wrf_in_prod
                    break

    def make_gsi_namelist(self,filename='gsiparm.anl'):
        """Creates the GSI namelist in the specified file."""
        logger=self.log()
        invars=dict()

        if self.wrf_top_Pa < 201.0 and self.confbool('sat_da',True):
            invars.update( USE_GFS_STRATOSPHERE=True,
                           USE_GFS_OZONE=True,
                           REGIONAL_OZONE=True,
                           NDAT=80 )
        else:
            invars.update( USE_GFS_STRATOSPHERE=False,
                           USE_GFS_OZONE=False,
                           REGIONAL_OZONE=False,
                           NDAT=16 )

        if not self.confbool('use_newradbc',False):
            SETUP='upd_pred=0,'
        else:
            SETUP='''newpc4pred=.true., adp_anglebc=.true., angord=4
                     passive_bc=.false., use_edges=.false., emiss_bc=.true.
                     diag_precon=.true., step_start=1.e-3,upd_pred=0,'''
        dom=self._sim[self._domain]
        nx=dom.nx
        assert(nx)
        ny=dom.ny
        assert(ny)
        nz=dom.nz
        assert(nz)
        invars.update(NLAT=ny-1, NLON=nx-1, NETCDF=True, LEVS=nz-1,
                      NLON_ENS_REGIONAL=nx-1,SETUP=SETUP,
                      NLAT_ENS_REGIONAL=ny-1)
        invars.update(tdrtype='tldplrbufr')#FIXME: tdrtype
        invars.update(gps_dtype='gps_bnd')

        nml_file=self.confstr('nml_file')
        nml_section=self.confstr('nml_section',self.section)
        atime=self.atime
        ni=hwrf.namelist.NamelistInserter(self.conf,nml_section)

        with open(nml_file,'rt') as nf:
            with open(filename,'wt') as of:
                of.write(ni.parse(nf,logger=logger,source=nml_file,
                                  raise_all=True,atime=atime,**invars))

    def after_gsi(self):
        """This is intended to be overridden by subclasses to perform
        some action after gsi is complete, but before products are
        delivered.  The default implementation does nothing."""

    def before_gsi(self):
        """This is intended to be overridden by subclasses to perform
        some action after all inputs needed for gsi are available, but
        before gsi starts.  The default implementation does nothing."""

    def grab_more_inputs(self):
        """This is intended to be overridden by subclasses to copy or
        link more inputs for GSI.  The default implementation does
        nothing."""

    def grab_bias_satang(self):
        """Copies or links bias correction and satellite angle files"""
        logger=self.log()
        #parent_atime=self._atime
        parent_atime=self.parent_atime
        def get(ds,it,fn):
            logger.info('%s: search for this at dataset=%s item=%s time=%s'%(fn,ds,it,parent_atime.strftime('%Y%m%d%H')))
            there=self._in_catalog.locate(ds,it,atime=parent_atime,logger=logger,ftime=parent_atime)
            logger.info('%s: found at %s'%(fn,there))
            make_symlink(there,fn,force=True,logger=logger)
        if self.confbool('use_newradbc',False):
            get(self._biascr_dataset,self._biascr_item,'satbias')
            get(self._biascr_dataset,self._biascr_pc_item,'satbias_pc')
        else:
            get(self._biascr_dataset,self._abias_item,'satbias_in')
            get(self._biascr_dataset,self._satang_item,'satbias_angle')

    def run_gsi_exe(self):
        """Runs the actual GSI executable."""
        logger=self.log()
        cmd= ( mpirun(mpi(self.getexe('gsi')),allranks=True) < 'gsiparm.anl' )
        if self.redirect: cmd = cmd > 'stdout'
        logger.warning(repr(cmd))
        # Send a message at the highest logging level before and after
        # GSI so the NCEP-wide jlogfile has a message about the status
        # of GSI.
        self.postmsg('Starting GSI for %s domain'%(str(self.domain),))

        sleeptime=self.conffloat('sleeptime',30.0)

        try:
            checkrun(cmd,sleeptime=sleeptime)
        except Exception as e:
            logger.critical('GSI failed for %s domain: %s'
                            %(str(self.domain),str(e)),exc_info=True)
            raise
        self.postmsg('GSI succeeded for %s domain'%(str(self.domain),))

    def deliver_products(self):
        """Delivers output products to the intercom directory."""
        d=os.path.dirname(self._wrf_out_prod.location)
        logger=self.log()
        produtil.fileop.makedirs(d,logger=logger)
        self._wrf_out_prod.deliver(frominfo='wrf_inout',keep=False,
                                   logger=logger)
        produtil.fileop.deliver_file(
            'satbias_out',os.path.join(self.outdir,'satbias_out'),
            keep=False,logger=logger)

    def run(self):
        """Runs the GSI and delivers the results."""
        logger=self.log()
        dirname=self.workdir
        logger.info('Run gsi in directory %s'%(dirname,))
        if os.path.exists(dirname):
            logger.info('Delete old data in %s'%(dirname,))
            shutil.rmtree(dirname)
        with NamedDir(dirname):
            self.grab_fix_parm()
            self.grab_gfs_enkf()
            self.grab_bufr()
            self.grab_prepbufr()
            self.grab_bias_satang()
            self.grab_more_inputs()
            self.copy_wrf_inout()
            self.make_gsi_namelist()

            self.before_gsi()
            self.run_gsi_exe()
            self.after_gsi()

            self.deliver_products()
        self.state=COMPLETED

    def grab_fix_parm(self):
        """Links or copies to the local directory any fix or parm
        files needed by GSI."""
        logger=self.log()
        logger.info("Copying fix and parm files")
        s=self.icstr
        def lnsf(a,b):
            produtil.fileop.make_symlink(a,b,logger=logger,force=True)

        anavinfo=s('{FIXgsi}/anavinfo_hwrf_d2')
        berror=s('{FIXgsi}/nam_glb_berror.f77.gcv')
        emiscoef_IRwater=s('{FIXcrtm}/EmisCoeff/IR_Water/Big_Endian/Nalli.IRwater.EmisCoeff.bin')
        emiscoef_IRice=s('{FIXcrtm}/EmisCoeff/IR_Ice/SEcategory/Big_Endian/NPOESS.IRice.EmisCoeff.bin')
        emiscoef_IRland=s('{FIXcrtm}/EmisCoeff/IR_Land/SEcategory/Big_Endian/NPOESS.IRland.EmisCoeff.bin')
        emiscoef_IRsnow=s('{FIXcrtm}/EmisCoeff/IR_Snow/SEcategory/Big_Endian/NPOESS.IRsnow.EmisCoeff.bin')
        emiscoef_VISice=s('{FIXcrtm}/EmisCoeff/VIS_Ice/SEcategory/Big_Endian/NPOESS.VISice.EmisCoeff.bin')
        emiscoef_VISland=s('{FIXcrtm}/EmisCoeff/VIS_Land/SEcategory/Big_Endian/NPOESS.VISland.EmisCoeff.bin')
        emiscoef_VISsnow=s('{FIXcrtm}/EmisCoeff/VIS_Snow/SEcategory/Big_Endian/NPOESS.VISsnow.EmisCoeff.bin')
        emiscoef_VISwater=s('{FIXcrtm}/EmisCoeff/VIS_Water/SEcategory/Big_Endian/NPOESS.VISwater.EmisCoeff.bin')
        emiscoef_MWwater=s('{FIXcrtm}/EmisCoeff/MW_Water/Big_Endian/FASTEM5.MWwater.EmisCoeff.bin')
        aercoef=s('{FIXcrtm}/AerosolCoeff/Big_Endian/AerosolCoeff.bin')
        cldcoef=s('{FIXcrtm}/CloudCoeff/Big_Endian/CloudCoeff.bin')
        satinfo=s('{FIXgsi}/hwrf_satinfo.txt')
        atmsfilter=s('{FIXgsi}/atms_beamwidth.txt')
        scaninfo=s('{FIXgsi}/global_scaninfo.txt')
        satangl=s('{FIXgsi}/nam_global_satangbias.txt')
        pcpinfo=s('{FIXgsi}/nam_global_pcpinfo.txt')
        ozinfo=s('{FIXgsi}/global_ozinfo.txt')
        errtable=s('{FIXgsi}/hwrf_nam_errtable.r3dv')
        convinfo=s('{FIXgsi}/hwrf_convinfo.txt')
        
        # Only need this file for single obs test
        bufrtable=s('{FIXgsi}/prepobs_prep.bufrtable')
        
        # Only need this file for sst retrieval
        bftab_sst=s('{FIXgsi}/bufrtab.012')
        
        # NOTE: GSI now run for all storm depths.
        lnsf(anavinfo, './anavinfo')
        lnsf(berror,   './berror_stats')
        lnsf(emiscoef_IRwater, './Nalli.IRwater.EmisCoeff.bin')
        lnsf(emiscoef_IRice, './NPOESS.IRice.EmisCoeff.bin')
        lnsf(emiscoef_IRsnow, './NPOESS.IRsnow.EmisCoeff.bin')
        lnsf(emiscoef_IRland, './NPOESS.IRland.EmisCoeff.bin')
        lnsf(emiscoef_VISice, './NPOESS.VISice.EmisCoeff.bin')
        lnsf(emiscoef_VISland, './NPOESS.VISland.EmisCoeff.bin')
        lnsf(emiscoef_VISsnow, './NPOESS.VISsnow.EmisCoeff.bin')
        lnsf(emiscoef_VISwater, './NPOESS.VISwater.EmisCoeff.bin')
        lnsf(emiscoef_MWwater, './FASTEM5.MWwater.EmisCoeff.bin')
        lnsf(aercoef,  './AerosolCoeff.bin')
        lnsf(cldcoef,  './CloudCoeff.bin')
        lnsf(satangl,  './satbias_angle')
        lnsf(satinfo,  './satinfo')
        lnsf(scaninfo, './scaninfo')
        lnsf(pcpinfo,  './pcpinfo')
        lnsf(ozinfo,   './ozinfo')
        lnsf(convinfo, './convinfo')
        lnsf(errtable, './errtable')
        lnsf(atmsfilter, './atms_beamwidth.txt')
        lnsf(bufrtable, './prepobs_prep.bufrtable')
        lnsf(bftab_sst, './bftab_sstphr')
        
        # Copy CRTM coefficient files based on entries in satinfo file
        with open('satinfo','rt') as f:
            for line in f:
                splat=line.split()
                if not splat or splat[0][0]=='!': continue
                satsen=splat[0]
                spccoeff=s('{satsen}.SpcCoeff.bin',satsen=satsen)
                if not isnonempty(spccoeff):
                    make_symlinks_in([
                            s('{FIXcrtm}/SpcCoeff/Big_Endian/{spccoeff}',spccoeff=spccoeff),
                            s('{FIXcrtm}/TauCoeff/Big_Endian/{satsen}.TauCoeff.bin',satsen=satsen) ],
                                     '.',force=True,logger=logger)

class FGATGSI(GSIBase):
    def __init__(self,dstore,conf,section,domain,wrf_in_prod,fgat_in_prods,
                 sim,cycling_interval=None,taskname=None,atime=None,
                 **kwargs):
        """The GSIBase constructor:
           dstore -- passed to Datum: the Datastore object for this Task
           conf -- the conf object for this task (passed to HWRFTask)
           section -- the conf section for this task (passed to HWRFTask)
           domain -- the WRFDomain for this GSI.  Must have been
              initialized by a WRFSimulation
           wrf_in_prod -- the product that should be used for the
              analysis background
           fgat_in_prods -- a mapping from analysis time to a product
             for all of the FGAT times
           taskname -- Optional: the taskname for this product.
             Passed to HWRFTask
           atime -- the analysis time for this GSI.  Default: conf.cycle
           kwargs -- ignored; passed to HWRFTask"""
        assert(wrf_in_prod is not None)
        if cycling_interval is None: cycling_interval=6*3600
        self.cycling_interval=to_timedelta(cycling_interval)
        if atime is None: atime=conf.cycle
        for prod in fgat_in_prods.itervalues():
            assert(isinstance(prod,produtil.datastore.Product))
        self._fgat_in_prod = [ (prod,to_datetime_rel(time,atime)) \
            for time,prod in fgat_in_prods.iteritems() ]
        assert(wrf_in_prod is not None)
        super(FGATGSI,self).__init__(dstore,conf,section,domain,wrf_in_prod,
           sim,taskname=taskname,atime=atime,**kwargs)

    def parent_fhrs(self):
        """Iterates over all FGAT forecast hours (as ints) relative to
        the parent analysis time.  For example, 3, 6, and 9 for
        three-hourly FGAT."""
        parent_atime=self.parent_atime
        for p,t in self._fgat_in_prod:
            yield int(round(to_fraction(t-parent_atime)/3600))

    def fgat_fhrs(self):
        """Iterates over all FGAT forecast hours (as ints) relative to
        the parent analysis time.  For example, -3, 0, and 3 for
        three-hourly FGAT."""
        atime=self.atime
        for p,t in self._fgat_in_prod:
            yield int(round(to_fraction(t-atime,negok=True)/3600))

    def grab_more_inputs(self):
        """Links to the current working directory gdas native spectral
        output files for all FGAT hours."""
        atime=self.parent_atime
        logger=self.log()
        for hr in self.parent_fhrs():
            ftime = atime + datetime.timedelta(hours=hr)
            there=self._in_catalog.locate(self._sigma_dataset,self._sigma_item,
                atime=atime,ftime=ftime,logger=logger)
            if not isnonempty(there):
                raise GSIInputError(
                    '%s %s: required input file is empty or non-existant: %s'
                    %(self._sigma_dataset,self._sigma_item,there))
            here='gfs_sigf%02d'%(hr,)
            make_symlink(there,here,force=True,logger=logger)

    def copy_wrf_inout(self,filename='wrf_inout',others='wrf_inou%d'):
        """Copies the WRF analysis or input file to the specified
        filename.  Also copies the wrf input files for other FGAT
        hours to separate files specified by "others".  The "others"
        must be a format that includes at least one integer
        argument."""
        logger=self.log()
        atime=self.atime
        names=dict()
        fci=to_fraction(self.cycling_interval)
        for p,t in self._fgat_in_prod:
            assert(isinstance(p,produtil.datastore.Product))
            dt=int(round(float((to_fraction(t-atime,negok=True) + fci)/3600.0)))
            name=others%dt
            logger.debug('add prod=%s time=%s'%(repr(p.did),repr(dt)))
            names[p]=name

        names[self._wrf_in_prod]='wrf_inout'

        for p,n in names.iteritems():
            logger.debug('prod %s has name %s'%(repr(p.did),repr(n)))

        def namer(p,logger,*args): return names[p]
        def actor(p,name,logger,*args): 
            deliver_file(p.location,name,logger=logger)
        produtil.datastore.wait_for_products([a for a in names.iterkeys() ],
                                             logger,namer,actor)

