__all__=["DataCatalog","InputSource"]

import collections, os, ftplib, tempfile, ConfigParser, urlparse, stat, re, threading
import produtil.run, produtil.cluster, produtil.fileop
import hwrf.numerics, hwrf.exceptions

from produtil.run import alias, exe, checkrun, ExitStatusException
from produtil.fileop import deliver_file, isnonempty, make_symlink
from hwrf.numerics import to_datetime
from hwrf.exceptions import InputSourceBadType,PartialTransfer,UnsupportedTransfer

########################################################################
def tempopen(f,m):
    return tempfile.NamedTemporaryFile(prefix=os.path.basename(f),
                                  dir=os.path.dirname(f),
                                  mode=m,suffix='.tmp',delete=False)

########################################################################
class DataCatalog(object):
    """This class is a collection of functions that know how to
    provide the location of a file in either an archive or a
    filesystem.  It does not know how to actually obtain the file.
    This serves as the underlying "where is that file" implementation
    of InputSource."""
    def __init__(self,conf,section,anltime):
        self.conf=conf
        if not isinstance(section,basestring):
            raise TypeError('In DataCatalog.__init__, section must be a string.')
        self.section=section
        self.anltime=to_datetime(anltime)
    def __repr__(self):
        return "DataCatalog(%s,%s,%s)"%(repr(self.conf),repr(self.section),
                                        repr(self.anltime))
    def rt_updated(self):
        """Returns True if this dataset is updated in real-time, False
        otherwise.  By default, this will return True if
        conf[section,"rt_updated"] is set to "yes" or False otherwise."""
        try:
            return conf.getbool(section,'rt_updated',False)
        except ( ConfigParser.Error,KeyError,TypeError,ValueError ) as e:
            return False
    def parse(self,string,atime=None,ftime=None,logger=None,**kwargs):
        """Performs string interpolation using the underlying conf
        object.  This acts exactly like the expansions done in the
        hwrf.conf file: {stuff} is expanded to the contents of the
        "stuff" variable.  Expanstions are done in the section
        specified in the constructor.  In addition, various a* and f*
        variables are expanded based on the analysis time ("atime")
        and forecast time ("ftime").  See
        hwrf.HWRFConfig.timestrinterp for details."""
        if atime is None: 
            if logger is not None:
                logger.info(
                    '{%s}: has no atime.  Will use atime=self.anltime=%s.'%(
                        str(string),repr(atime)))
            atime=self.anltime
        if ftime is None: 
            if logger is not None:
                logger.info('{%s}: has no ftime.  Will use ftime=atime=%s.'%(
                        str(string),repr(atime)))
            ftime=atime
        if logger is not None:
            logger.info(
                'parsing {%s} with ftime=%s atime=%s in section %s'
                %(str(string),repr(ftime),repr(atime),repr(self.section)))
        return self.conf.timestrinterp(
            self.section,"{"+string+"}",ftime,atime,**kwargs)
    def locate(self,dataset,item,atime=None,ftime=None,logger=None,**kwargs):
        """Locates the specified item for the specified dataset, at
        the given analysis time ("atime") and forecast time ("ftime").
        If the requested data is known to not exist, returns None.
        This should be overridden by subclasses.  The present
        implementation just does this: {dataset}/{item} Expanding
        dataset and item with self.parse.  Any kwargs are passed
        along: this allows such things as ensemble ID, or switching
        between GRIB1 or GRIB2 via a keyword argument."""
        if logger is not None: 
            logger.info(
                'locate item=%s atime=%s ftime=%s in dataset=%s'
                %(repr(item),repr(atime),repr(ftime),repr(dataset)))
        ds=self.parse(dataset,atime=atime,ftime=ftime,logger=logger,**kwargs)
        it=self.parse(item,atime=atime,ftime=ftime,logger=logger,**kwargs)
        result=urlparse.urljoin(ds,it)
        if logger is not None:
            logger.info( 'result is %s'%(repr(result),))
        return result

########################################################################
year2010start=to_datetime('2010010100')
class ZeusCatalog(DataCatalog):
    """This subclass of DataCatalog knows to refuse pre-2010 data
    (which is nowhere on Zeus)"""
    def locate(self,dataset,item,atime=None,ftime=None,**kwargs):
        """Locates the specified item for the specified dataset, at
        the given analysis time ("atime") and forecast time ("ftime").
        If the requested data is known to not exist, returns None.
        This should be overridden by subclasses.  The present
        implementation just does this: {dataset}/{item} Expanding
        dataset and item with self.parse.  Any kwargs are passed
        along: this allows such things as ensemble ID, or switching
        between GRIB1 or GRIB2 via a keyword argument."""
        if atime is None: atime=self.anltime
        if atime<year2010start: return None
        return urlparse.urljoin(self.parse(dataset,atime,ftime,**kwargs),
                                self.parse(item,atime,ftime,**kwargs))

########################################################################
class InputSource(object):
    """This class knows how to fetch data from remote clusters, or the
    local machine.  The data locations are specified by a DataCatalog,
    and the transfer mechanisms are specified in a config file."""
    def __init__(self,conf,section,anltime):
        self.conf=conf
        self.section=section
        self.anltime=anltime
        self.forecast=list() # FORECAST mode DataCatalogs
        self._f_sorted=True
        self.history=list() # HISTORY mode DataCatalogs
        self._h_sorted=True

        sections=[section]
        if conf.has_option(section,'@inc'):
            sections.extend(conf[section,'@inc'].split(','))

        sources=collections.defaultdict(dict)
        for sec in sections:
            for key in conf.keys(sec):
                c=key.find('%')
                if(c>0):
                    (src,attr)=(key[0:c],key[c+1:])
                    sources[src][attr]=conf.get(sec,key)
        for (src,attr) in sources.iteritems():
            if 'location' in attr and ('histprio' in attr or \
                                           'fcstprio' in attr):
                dctype=attr.get('type','DataCatalog')
                if   dctype=='DataCatalog': 
                    dc=DataCatalog(self.conf,src,self.anltime)
                elif dctype=='ZeusCatalog':  
                    dc=ZeusCatalog(self.conf,src,self.anltime)
                else:
                    raise InputSourceBadType(
                        'Do not know how to make a DataCatalog of type "%s"'
                        %(dctype,))
                self.add(dc,location=attr['location'],
                         fcstprio=attr.get('fcstprio',None),
                         histprio=attr.get('histprio',None))
        self._sort()
    def _rsync_ssh_exe(self,netpart,path,dest):
        """Returns a Runner object (as in produtil.run) for executing
        rsync -e ssh.  This subroutine is used to implement
        workarounds for known bugs."""
        if 'jet' in netpart or produtil.cluster.name()=='jet':
            # Workaround for Jet bug: use protocol 29
            return alias(exe('rsync')[
                    '-e','ssh','--protocol','29','-lvptgoD',
                    "%s:%s"%(netpart,path),str(dest)])
        else:
            return alias(exe('rsync')[
                    '-e','ssh','-lvptgoD',"%s:%s"%(netpart,path),dest])
    def _sort(self):
        """Sorts the list of history and forecast DataCatalogs by
        decreasing priority."""
        self.forecast=sorted(self.forecast,key=lambda x: -x[0])
        self.history=sorted(self.history,key=lambda x: -x[0])
    def add(self,dc,location,fcstprio=None,histprio=None):
        """Adds a DataCatalog to this InputSource.  Note that bad
        things will happen if you add the same source twice

        Arguments:
          dc = the DataCatelog object
          location = the URL of the data source, including the
            username if needed.
            Examples:
              local files: file:///
              scp:         sftp://Some.Username@dtn-zeus.rdhpcs.noaa.gov/
              ftp:         ftp://anonymous@ftpprd.ncep.noaa.gov/
          fcstprio = the priority for using this source in FORECAST
            (real-time) mode.  If missing or None, the source will not
            be used in FORECAST mode.

          histprio = the priority for using this source in HISTORY
            (retrospective) mode.  If missing or None,the source will
            not be used in HISTORY mode.
            
          If fcstprio and histprio are both None, the request is
          ignored."""
        if fcstprio is None and histprio is None: return
        parsed=urlparse.urlparse(location)
        if fcstprio is not None:
            self.forecast.append( ( int(fcstprio), location, parsed, dc ) )
            self._f_sorted=False
        if histprio is not None:
            self.history.append( ( int(histprio), location, parsed, dc ) )
            self._h_sorted=False
    def open_ftp(self,netpart,logger=None,timeout=20):
        """Opens the specified ftp://user@host/... request subject to
        the specified timeout, logging to the specified logger (if
        present and non-Null)."""
        if logger is not None:
            logger.info('open_ftp %s'%(netpart,))
        r=re.search('([a-zA-Z0-9_.-]+)+@(.+)',netpart)
        if r:
            (user,host)=r.groups()
            if not user or not host:
                raise InvalidLogin(
                    'FTP logins must be of the form user@host but you '
                    'gave "%s"'%(netpart))
        else:
            (user,host)=('anonymous',netpart)
        f=None
        try:
            if logger is not None: logger.info('%s@%s: log in'%(user,host))
            f=ftplib.FTP(host,user,timeout=timeout)
            f.login()
            assert(f is not None)
            retval=f
            f=None
            return retval
        finally:
            if f is not None: f.close()
    def fetch_htar(self,dc,dsurl,urlmore,logger=None):
        """Placeholder for future development: raises
        NotImplementedError.  This will fetch a file from an htar
        archive."""
        raise NotImplementedError('HTAR support is not implemented yet')
    def fetch_file(self,streams,dc,dsurl,urlmore,dest,logger=None,
                   timeout=20,realtime=True):
        """You should not call this directly; it is meant to be called
        by "get" and re-implemented in subclasses.  This grabs one
        file, potentially from a remote location.  The URL for the
        base directory of some dataset is in dsurl, while the specific
        file is in urlmore.  The urlmore will be appended to the file
        part of dsurl via urljoin, and the resulting file will be
        transferred."""
        parsed=urlparse.urlparse(dsurl)
        if parsed.scheme == 'htar':
            # Target is an htar archive
            return self.fetch_htar(self,dc,dsurl,urlmore,logger=logger)
        joined=urlparse.urljoin(dsurl,urlmore,allow_fragments=True)
        parsed=urlparse.urlparse(joined)
        if logger is not None:
            logger.info('%s + %s = %s',repr(dsurl),repr(urlmore),repr(joined))
        scheme=parsed.scheme
        path=parsed.path
        netpart=parsed.netloc
        if scheme== 'file':
            if logger is not None:
                logger.info('%s: from local file %s'%(dest,joined))
            if ( realtime and dc.rt_updated() ) or os.path.exists(path):
                make_symlink(path,dest,force=True,logger=logger)
            else:
                return False
                #produtil.fileop.deliver_file(path,dest,keep=True,logger=logger)
        elif scheme=='ftp':
            n="%s://%s"%(scheme,netpart)
            if n not in streams:
                streams[n]=self.open_ftp(n,logger=logger,timeout=timeout)
            stream=streams[n]
            tempname=None
            try:
                with tempopen(dest,'wb') as f:
                    tempname=f.name
                    if logger is not None:
                        logger.info('%s: pull %s => %s'
                                    %(n,parsed.path,tempname))
                    stream.retrbinary("RETR "+parsed.path,f.write)
                remote_size=stream.size(parsed.path)
                if remote_size is not None:
                    local_size=os.path.getsize(tempname)
                    if local_size!=remote_size:
                        if logger is not None:
                            logger.warning(
                                '%s: wrong size: %d local vs %d remote'
                                %(tempname,local_size,remote_size))
                        raise PartialTransfer(
                            'Could not transfer full file: only %d of %d '
                            'bytes transferred.'%(local_size,remote_size))
                if logger is not None:
                    logger.info('%s: move from %s'%(dest,tempname))
                os.rename(tempname,dest)
                tempname=None
            finally:
                if tempname is not None: os.remove(tempname)
        elif scheme=='sftp':
            tempname=None
            try:
                with tempopen(dest,'wb') as f:
                    tempname=f.name
                cmd=self._rsync_ssh_exe(netpart,path,tempname)
                checkrun(cmd,logger=logger)
                os.rename(tempname,dest)
                tempname=None
            finally:
                if tempname is not None: os.remove(tempname)
        else:
            raise UnsupportedTransfer(
                'Cannot transfer this url: unsupported method (not htar, '
                'ftp, file or sftp): '+joined)
        return True
    def list_for(realtime=True):
        if realtime:
            if not self._f_sorted: self._sort()
            return self.forecast
        else:
            if not self._h_sorted: self._sort()
            return self.history
    def get(self,data,target_dc,realtime=False,logger=None,
            skip_existing=True):
        """Transfers the specified set of data to the specified
        target.  The "target_dc" is a DataCatalog that specifies the
        destination filenames.  The "realtime" argument is True for
        FORECAST (real-time) mode runs, and False for HISTORY
        (retrospective) mode runs.  The "data" argument should be an
        iterable (list, tuple, etc.) where each element is a dict-like
        object that describes one file to obtain.  Each dict contains:

          dataset - string name of the dataset (GFS_history, GDAS1,
            GFS_ENKF, etc.)
          item - string name of the object (ie.: sf, sfcanl, bufr)
          atime - Optional: a datetime.datetime specifying the
            analysis time.  Default is the atime from the
            InputSource's constructor.
          ftime - Optional: a datetime.datetime specifying the
            forecast time.
          others - any other keyword arguments will be sent to the
            .location functions in any of this InputSource's
            DataCatalog objects."""  
        dclist=self.list_for(realtime)
        done=set()
        for ( prio, loc, parsed, dc ) in dclist:
            assert(loc is not None)
            assert(prio is not None)
            assert(parsed is not None)
            assert(dc is not None)
            streams=dict()
            archives=collections.defaultdict(dict)
            try:
                i=0
                for d in data:
                    i+=1
                    if i in done: continue # skip files we already
                                           # transferred
                    tgt=target_dc.locate(**d)
                    if tgt is None:
                        raise InputSourceNoTarget(
                            'Input source %s provided no target location '
                            'for %s'%(repr(tgt),repr(d)))
                    if os.path.exists(tgt) and skip_existing:
                        if logger is not None: 
                            logger.info('%s: already exists'%(tgt,))
                        done.add(i)
                        continue
                    if logger is not None:
                        logger.info('search for %s in %s'%(repr(d),repr(dc)))
                    src=dc.locate(**d)
                    if src is None: continue
                    archsep=src.find('#')
                    if archsep>=0:
                        # This is in an archive, so we will have to stage
                        # the archive first, and get the file in the
                        # second pass.
                        arch=src[0:archsep]
                        filepart=src[archsep+1:]
                        archives[arch][filepart]=tgt
                    else:
                        try:
                            if self.fetch_file(
                                streams,dc,loc,src,tgt,
                                logger=logger,realtime=realtime):
                                done.add(i)
                        except (EnvironmentError,ExitStatusException) as e:
                            if logger is not None:
                                logger.warning(
                                    'fetching %s=>%s: %s'%(str(src),str(tgt),
                                        str(e)),exc_info=True)
                # FIXME: ADD SECOND PASS TO GRAB ARCHIVES AND
                # DE-ARCHIVE ARCHIVED ITEMS
                if False:
                    for (archpath,parts) in archives:
                        # FIXME: STAGE ARCHIVE HERE
                        for (filepart,tgt) in parts:
                            deliver_file(os.path.join(stagepath,filepart),
                                         tgt,keep=False)
                        # FIXME: DELETE STAGED DATA HERE
            finally:
                for (key,stream) in streams.iteritems(): 
                    try:
                        stream.close()
                    except Exception as e:
                        if logger is not None:
                            logger.warning(
                                'Exception while closing stream %s: %s'
                                %(key,str(e)),exc_info=True)
            if archives:
                raise UnsupportedTransfer(
                    'De-archiving is not supported.  Use file, sftp or '
                    'ftp transfers instead.')
    def get_one(self,dataset,item,dest,logger=None,timeout=20,realtime=True,
                **kwargs):
        """This is a simple wrapper around fetch_file that gets only
        one file."""
        streams=dict()
        try:
            dclist=list_for(realtime)
            for ( prio, loc, parsed, dc ) in dclist:
                src=dc.locate(dataset=dataset,item=item,**kwargs)
                if src is None: continue
                archsep=src.find('#')
                if archsep>=0:
                    #FIXME: IMPLEMENT DE-ARCHIVING
                    raise NotImplementedError(
                        'Source is in an archive.  De-archiving is not '
                        'supported yet.')
                elif self.fetch_file(streams,dc,loc,src,dest,logger=logger): 
                    break
        finally:
            for (key,stream) in streams.iteritems(): 
                try:
                    stream.close()
                except Exception as e:
                    if logger is not None:
                        logger.warning(
                            'Exception while closing stream %s: %s'
                            %(key,str(e)),exc_info=True)
