"""This module is used to create the initial HWRF conf file in the
first HWRF job via the hwrf.launcher.launch.  The hwrf.launcher.load
then reloads that configuration.  The launch() function does more than
just create the conf file though.  It parses the tcvitals, creates
several initial files and directories and runs a sanity check on the
whole setup."""

__all__=['load','launch','HWRFLauncher']

import os
import produtil.fileop
import hwrf.revital, hwrf.storminfo, hwrf.numerics, hwrf.config, hwrf.input

from produtil.fileop import isnonempty
from hwrf.numerics import to_datetime_rel, to_datetime
from hwrf.config import HWRFConfig
from hwrf.exceptions import HWRFDirInsane,HWRFStormInsane,HWRFCycleInsane, \
    HWRFVariableInsane,HWRFInputInsane

def load(filename):
    """Creates an HWRFConfig object for an HWRF workflow that was
    previously initialized by hwrf.launcher.launch.  The only argument
    is the name of the config file produced by the launch command."""
    conf=HWRFLauncher()
    conf.read(filename)
    logger=conf.log()

    strcycle=conf.getstr('config','cycle')
    cycle=to_datetime(strcycle)
    conf.cycle=cycle
    logger.info('Running cycle: '+cycle.strftime('%Y%m%d%H'))
    
    WORKhwrf=conf.getdir('WORKhwrf')

    tmpvit=os.path.join(WORKhwrf,'tmpvit')
    logger.info(tmpvit+': read vitals for current cycle')
    with open(tmpvit,'rt') as f:
        syndat=hwrf.storminfo.parse_tcvitals(f,logger,raise_all=True)
        syndat=syndat[0]
    logger.info('Current cycle vitals: '+syndat.as_tcvitals())

    oldvit=os.path.join(WORKhwrf,'oldvit')
    logger.info(oldvit+': read vitals for prior cycle')
    with open(oldvit,'rt') as f:
        oldsyndat=hwrf.storminfo.parse_tcvitals(f,logger,raise_all=True)
        oldsyndat=oldsyndat[0]
    logger.info('Prior cycle vitals: '+oldsyndat.as_tcvitals())

    conf.set_storm(syndat,oldsyndat)

    return conf

def launch(file_list,cycle,stid=None,moreopt=None):
    """Initializes the directory structure for a new HWRF workflow.
    Returns the full path to the conf file that is created as a
    result.  That conf file should be passed in to the
    hwrf.launcher.load at the top of every job.

    file_list - a list of conf files to read
    cycle - the cycle to run; anything accepted by to_datetime
    stid - the three character storm identifier for the storm to run.
      For example, stid=11L is the eleventh storm of the season in the
      Atlantic basin.  Although this argument is optional, the single
      storm HWRF workflow will fail if stid is not provided.

    moreopt - a dict of dicts with additional options to set.  This
      maps section name to option to value.
    """
    for filename in file_list:
        if not isinstance(filename,basestring):
            raise TypeError('First input to hwrf.config.for_initial_job '
                            'must be a list of strings.')
    conf=HWRFLauncher()
    logger=conf.log()
    conf.cycle=to_datetime(cycle)
    for filename in file_list:
        logger.info("%s: parse this file"%(filename,))
        conf.read(filename)
    if moreopt is not None:
        for section,options in moreopt.iteritems():
            if not conf.has_section(section):
                conf.add_section(section)
            for option,value in options.iteritems():
                logger.info('Override: %s.%s=%s'
                            %(section,option,repr(value)))
                conf.set(section,option,value)
    conf.guess_default_values()
    cycling_interval=conf.getfloat('config','cycling_interval',6.0)
    cycling_interval=-abs(cycling_interval*3600.0)
    other_cycle=to_datetime_rel(cycling_interval,conf.cycle)

    if stid is not None:
        revit=conf.read_tcvitals_and_messages(other_cycle=other_cycle)
        conf.gen_vitals(stid,cycling_interval,revit)

    produtil.fileop.makedirs(conf.getdir('com'),logger=logger)
    produtil.fileop.makedirs(conf.getdir('WORKhwrf'),logger=logger)

    logger.info('Expand certain [dir] values to ensure availability '
                'before vitals parsing.')
    for var in ( 'WORKhwrf', 'HOMEhwrf', 'com' ):
        expand=conf.getstr('dir',var)
        logger.info('Replace [dir] %s with %s'%(var,expand))
        conf.set('dir',var,expand)

    if stid is not None:
        conf.decide_domain_center()
        loc=conf.getdir('domlocfile')
        logger.info('%s: Writing domain center.'%(loc,))
        with open(loc,'wt') as f:
            f.write("%g\n%g\n"%(
                    conf.getfloat('config','domlat'),
                    conf.getfloat('config','domlon')))

    confloc=conf.getloc('CONFhwrf')
    logger.info('%s: write hwrf.conf here'%(confloc,))
    with open(confloc,'wt') as f:
        conf.write(f)

    with open(os.path.join(conf.getdir('WORKhwrf'),'PDY'),'wt') as f:
        f.write(conf.strinterp(
                'config','cyc={HH}\nPDY={YMD}\nYMDH={YMDH}\n'))

    return conf

class HWRFLauncher(HWRFConfig):
    def __init__(self,conf=None):
        super(HWRFLauncher,self).__init__(conf)

    def storm_for_stormnum(self):
        """Returns the one letter basin, numeric storm ID and year for
        the specified storm number (1-10)."""

    def decide_domain_center(self,logger=None):
        """If the domain center is not already set in the [config]
        section domlat and domlon variables, decides the domain center
        using the hwrf.storminfo.StormInfo.hwrf_domain_center
        routine."""
        if logger is None: logger=self.log()
        if self.has_option('config','domlat') and \
                self.has_option('config','domlon'):
            cenla=self.getfloat('config','domlat')
            cenlo=self.getfloat('config','domlon')
            logger.info('Domain center is already set to lat=%g lon=%'
                        %(cenla,cenlo))
            return
        (cenlo, cenla) = self.syndat.hwrf_domain_center(logger)
        self.set('config','domlat',cenla)
        self.set('config','domlon',cenlo)
        logger.info('Decided on domain center lat=%g lon=%g'%(cenla,cenlo))

    def write_vitals(self,vitlist=None):
        """Writes vitals files to the expected locations in WORKhwrf.
        The optional vitlist argument is a list or tuple of storm
        numbers (1-10) whose vitals should be written."""
        if vitlist is None: 
            storm_num=self.getint('config','storm_num',1)

    def choose_vitbase(self,storm_num=None):
        """Decides the location of the vitfile that should be read in
        by read_precleaned_vitfile.  Optionally, you can specify the
        storm number (1-10) of the storm whose vitals should be read
        in.  Otherwise, a reasonable guess will be made."""
        if storm_num is not None:
            storm_num=int(storm_num)
            vitfile=os.path.join(self.getdir('WORKhwrf'),
                'storm%d.vitals'%(storm_num,))
        else:
            stormlabel=self.getstr('config','stormlabel','storm1')
            vitfile=os.path.join(self.getdir('WORKhwrf'),
                '%s.vitals'%(stormlabel,))
        return vitfile

    def _set_auto_vars(self,HOMEhwrf=None,WORKhwrf=None):
        """FIXME: DELETE THIS FUNCTION"""
        if not self._conf.has_option('dir','HOMEhwrf'):
            self._conf.set('dir','HOMEhwrf',HOMEhwrf)
        if not self._conf.has_option('dir','WORKhwrf'):
            self._conf.set('dir','WORKhwrf',WORKhwrf)
        if not self._conf.has_option('config','datastore'):
            self._conf.set('config','datastore',
                           '{WORKhwrf}/hwrf_state.sqlite3')

    def read_tcvitals_and_messages(self,vitdir=None,vitpattern=None,
            include_messages=True,other_cycle=None):
        """Reads in the tcvitals files for the current cycle and
        optionally another cycle, which may be in the same file.  Also
        reads in message files if requested.  Cleans the result up and
        returns it as an hwrf.revital.Revital object.

           vitdir - optional: the directory in which to find the tcvitals.
              Default: [dir] section syndat variable.

           vitpattern - optional: passed into strftime to generate the
              name of the vitals file within vitdir. Default: [conf]
              section vitpattern variable, or syndat_tcvitals.%Y if
              missing.

           include_messages - optional flag: if True, attempts to find
              the hurricane message files, and includes them in the
              list of files to read in.  Default: True.

           other_cycle - optional: another cycle whose vitals file
              should also be parsed.  This can be anything accepted by
              to_datetime_rel(...,self.cycle).  This is intended to
              allow year-crossing cycling, such as a January 1, 00:00
              UTC cycle that is a warm start off of a prior December
              31, 18:00 UTC cycle.  If the other_cycle's vitals file
              is the same as the one from self.cycle, then the file is
              only read once."""
        ENV=os.environ
        logger=self.log()
        inputs=list()
        if vitdir is None:        
            vitdir=self.getdir('syndat')
        if vitpattern is None:    
            vitpattern=self.getstr('config','vitpattern',
                                   'syndat_tcvitals.%Y')
        file1=os.path.join(vitdir,self._cycle.strftime(vitpattern))
        inputs.append(file1)
        if other_cycle is not None:
            other_cycle=to_datetime_rel(other_cycle,self._cycle)
            file2=os.path.join(vitdir,other_cycle.strftime(vitpattern))
            if file2!=file1: 
                inputs.append(file2)

        if include_messages:
            # Try to guess the location of the message files:
            mdir=self.getdir('mesagdir','')
            if mdir is None or mdir=='':
                if 'mesagdir' in ENV:
                    mdir=ENV['messagdir']
                elif 'envir' in ENV:
                    mdir='/com/hur/'+ENV['envir']+'/inpdata'
                else:
                    mdir='/com/hur/prod/inpdata'

                # Add the messages to the input files:
            for imessage in xrange(5):
                file=os.path.join(mdir,'message%d'%(imessage+1,))
                if os.path.exists(file):
                    inputs.append(file)

        self.log().info('read vitals from: '+','.join(inputs))
        revital=hwrf.revital.Revital(logger=logger)
        revital.readfiles(inputs)
        return revital

    def set_storm(self,syndat,oldsyndat):
        """Sets the syndat and oldsyndat member variables, and several
        related options in the [config] section, to the storm in the
        provided tcvitals or message file data.  

          config.STID -- The three character storm id (ie.: 12L) of
            the storm to run.

          config.stnum -- the numeric part of config.STID
          config.basin1 -- the basin part of STID (ie.: the L in 12L)
          config.basin1lc -- the lower-case version of config.basin1
          """
        assert(isinstance(syndat,hwrf.storminfo.StormInfo))
        if oldsyndat is not None:
            assert(isinstance(oldsyndat,hwrf.storminfo.StormInfo))
        self.set_options('config',STID=syndat.stormid3,stnum=syndat.stnum,
                         basin1=syndat.basin1,basin1lc=syndat.basin1lc)
        self.__dict__['syndat']=syndat.copy()
        if oldsyndat is not None:
            self.__dict__['oldsyndat']=oldsyndat.copy()

    def gen_vitals(self,STID,cycling_interval,revital,storm_num=None):
        """Given an hwrf.revital.Revital object, preferably from
        read_precleaned_vitfile or read_tcvitals_and_messages,
        searches for the specified storm's vitals.  Creates the files
        that are expected to exist in the WORKhwrf directory.  The
        filenames are based off of the vitbase variable, but with
        various suffixes appended.  This function should only be
        called once per workflow, per storm."""
        logger=self.log()
        stnum=int(STID[0:2],10)
        STID=STID.upper()
        strcycle=self._cycle.strftime('%Y%m%d%H')
        syndat=None
        oldsyndat=None

        if cycling_interval<0:
            cycling_interval=-cycling_interval
        prior=hwrf.numerics.to_datetime_rel(-cycling_interval,self._cycle)
        strprior=prior.strftime('%Y%m%d%H')
        logger.info('gen_vitals: cycle=%s interval=%s prior=%s STID=%s'%(
                repr(self.cycle),repr(cycling_interval),repr(prior),
                repr(STID)))

        def keep_condition(vit):
            return vit.stormid3.upper()==STID or \
                ( 'old_stormid3' in vit.__dict__ and 
                  vit.old_stormid3.upper()==STID )

        if stnum>=50:
            logger.info('%s: Not renumbering invests because %d>=50.'
                        %(STID,stnum))
            unrenumbered=revital.copy()
            unrenumbered.discard_except(keep_condition)
            unrenumbered.clean_up_vitals()
            renumbered=unrenumbered
        else:
            logger.info('%s: Renumber and unrenumber invests.'%(STID,))
            unrenumbered=revital.copy()
            unrenumbered.renumber(unrenumber=True)
            unrenumbered.discard_except(keep_condition)
            unrenumbered.clean_up_vitals()
            renumbered=unrenumbered.copy()
            renumbered.swap_numbers()
            renumbered.clean_up_vitals()
            unrenumbered.mirror_renumbered_vitals()
            unrenumbered.clean_up_vitals()

        # Find the current cycle's vitals:
        for vit in renumbered.each(STID):
            if vit.when==self._cycle:
                syndat=vit

        if syndat is None:
            raise hwrf.storminfo.NoSuchVitals(
                'Error: cannot find %s cycle %s'%(STID,strcycle))
        logger.info('syndat='+syndat.as_tcvitals())
        self.set_storm(syndat,None)

        # Find the prior cycle's vitals.  First pass: look for a cycle
        # whose data that actually exists on disk.
        nodatasyndat=None
        for vit in unrenumbered.each(STID,old=True):
            if vit.when!=prior: continue # wrong cycle
            if oldsyndat is not None and oldsyndat.stnum<50:
                logger.info('%s %s: not checking these vitals for data on '
                            'disk since I found a non-invest number %s '
                            'already with data on disk'
                            %(str(vit.stormid3),str(vit.YMDH),
                              str(oldsyndat.stormid3)))
            else:
                checkfile=self.timestrinterp(
                    'config','{HISTCHECK}',atime=prior,ftime=prior,
                    oldvit=vit.__dict__,vit=syndat.__dict__)
                if os.path.exists(checkfile):
                    logger.info('%s: exists'%(checkfile,))
                    logger.info('%s %s: prior is %s %s and has data on disk'%
                              (STID,strcycle,vit.stormid3,strprior))
                    oldsyndat=vit
                else:
                    logger.info('%s: does not exist'%(checkfile,))
                    logger.info('%s %s: prior could be %s %s but there is '
                                'no data on disk'%
                                (STID,strcycle,vit.stormid3,strprior))
            if oldsyndat is None:
                if nodatasyndat is not None and nodatasyndat.stnum<50:
                    logger.info('%s %s: not using as backup since I found a '
                                'non-invest number %s already'
                                %(str(vit.stormid3),str(vit.YMDH),
                                  str(nodatasyndat.stormid3)))
                else:
                    nodatasyndat=vit
                        
        self.set('config','expect_cold_start','no')
        if oldsyndat is None:
            logger.info('%s %s: no storm IDs for prior cycle have data '
                        'on disk.'%(STID,strcycle))
            if nodatasyndat is not None:
                oldsyndat=nodatasyndat
                logger.info('%s %s: will use %s %s as prior cycle storm.'
                          %(STID,strcycle,oldsyndat.stormid3,strprior))
                logger.info('prior vitals: '+oldsyndat.as_tcvitals())
            else:
                logger.warning('No prior syndat available.  This is a cold '
                               'start.  I will extrapolate vitals.')
                oldsyndat=syndat-cycling_interval # extrapolate vitals
                logger.warning('extrapolated vitals: %s'
                               %(oldsyndat.as_tcvitals()))
            self.set('config','expect_cold_start','yes')
        else:
            logger.info('%s %s prior cycle on disk for %s %s'
                        %(STID,strcycle,oldsyndat.stormid3,strprior))
            logger.info('prior cycle on disk: '+oldsyndat.as_tcvitals())

        self.set_storm(syndat,oldsyndat)
        vitbase=self.choose_vitbase(storm_num)

        vitbasedir=os.path.dirname(vitbase)
        produtil.fileop.makedirs(vitbasedir,logger=logger)

        logger.info('Reformat vitals...')
        filename=vitbase+'.allids'
        logger.info(
            filename+': write unrenumbered vitals with all storm IDs')
        with open(filename,'wt') as vitalsout:
            for vit in unrenumbered.each(stormid=STID,old=True):
                print>>vitalsout, vit.as_tcvitals()
        filename=vitbase+'.renumberlog'
        logger.info(filename+': write renumberlog with my storm ID')
        logger.info(vitbase+': write renumbered vitals')
        with open(filename,'wt') as renumberlog:
            with open(vitbase,'wt') as vitalsout:
                renumbered.print_vitals(vitalsout,renumberlog=renumberlog,
                                     stormid=STID,format='tcvitals')
        filename=vitbase+'.oldid'
        logger.info(filename+': write vitals with original ID')
        with open(filename,'wt') as vitalsout:
            for vit in renumbered.each(stormid=STID):
                print>>vitalsout, vit.old().as_tcvitals()

        filename=os.path.join(self.getdir('WORKhwrf'),'tmpvit')
        logger.info(filename+': write current cycle vitals here')
        with open(filename,'wt') as tmpvit:
            print>>tmpvit, self.syndat.as_tcvitals()

        filename=os.path.join(self.getdir('WORKhwrf'),'oldvit')
        logger.info(filename+': write prior cycle vitals here')
        with open(filename,'wt') as tmpvit:
            print>>tmpvit, self.oldsyndat.as_tcvitals()

    def sanity_check_directory(self,thedir,dirvar,writable=True,logger=None):
        """Checks to make sure the specified directory exists and can
        be read and executed.  If writable=True, also checks to see if
        it can be written.  The dirvar is an explanation of what the
        directory relates to, for example HOMEhwrf."""
        if logger is None: logger=self.log('sanity.checker')
        logger.info('%s: check directory %s'%(dirvar,thedir))
        if not os.path.exists(thedir):
            raise HWRFDirInsane('%s: directory does not exist: %s'
                                 %(dirvar,thedir),thedir)
        if writable:
            if not os.access(thedir,os.W_OK):
                raise HWRFDirInsane('%s: cannot write directory: %s'
                                    %(dirvar,thedir),thedir)
        if not os.access(thedir,os.R_OK):
            raise HWRFDirInsane('%s: cannot read directory: %s'
                                 %(dirvar,thedir),thedir)
        if not os.access(thedir,os.X_OK):
            raise HWRFDirInsane('%s: cannot execute directory: %s'
                                 %(dirvar,thedir),thedir)

    def sanity_check(self):
        """Runs simple sanity checks on the HWRF installation
        directory and configuration to make sure everything looks
        okay.  May throw a wide variety of exceptions if sanity checks
        fail."""
        logger=self.log('sanity.checker')
        for dirvar in ( 'WORKhwrf', 'com' ):
            logger.info('%s: check this dir variable'%(dirvar,))
            thedir=self.getdir(dirvar)
            self.sanity_check_directory(thedir,dirvar,True,logger)

        for dirvar in ( 'HOMEhwrf', 'EXEChwrf', 'EXhwrf', 'USHhwrf', 
                        'FIXhwrf', 'PARMhwrf', 'utilscript', 'utilexec' ):
            logger.info('%s: check this dir variable'%(dirvar,))
            thedir=self.getdir(dirvar)
            self.sanity_check_directory(thedir,dirvar,False,logger)

        CONFhwrf=self.getdir('CONFhwrf')
        logger.info('Try to load configuration file %s'%(CONFhwrf,))
        redo=load(CONFhwrf)

        logger.info('Compare new and old vitals')
        if 'syndat' in self.__dict__ and self.syndat.stormid3 != \
                redo.syndat.stormid3:
            raise HWRFStormInsane(
                "New directory has the wrong stormid: correct=%s conf=%s"
                %(self.syndat.stormid3,redo.syndat.stormid3))
        if self.cycle!=redo.cycle:
            raise HWRFCycleInsane(
                'New directory has the wrong cycle: correct=%s conf=%s'
                %(self.cycle.strftime('%Y%m%d%H'),
                  redo.cycle.strftime('%Y%m%d%H')))

        case_root=redo.getstr('config','case_root').upper()
        input_catalog=redo.getstr('config','input_catalog')
        logger.info('Case root is %s and input catalog is %s'
                    %(repr(case_root),repr(input_catalog)))
        if case_root=='HISTORY':
            logger.info('Retrospective mode.  Check for GFS sfcanl file.')
            ic=hwrf.input.DataCatalog(redo,input_catalog,redo.cycle)
            there=ic.locate('gfs','gfs_sfcanl',redo.cycle)
            if there is None:
                raise HWRFInputInsane('Could not locate gfs sfcanl file.')
            if not os.path.exists(there):
                raise HWRFInputInsane(
                    'GFS sfcanl file does not exist: '+there)
            if not isnonempty(there):
                raise HWRFInputInsane('GFS sfcanl file is empty: '+there)
        elif case_root=='FORECAST':
            logger.info('Real-time mode.  Will skip data checks.')
        else:
            raise HWRFVariableInsane(
                'config.case_root must be HISTORY or FORECAST not %s'
                %(repr(case_root),))

        self.sanity_check_expt(logger)

    def sanity_check_expt(self,logger):
        """Loads the hwrf_expt module, runs its init_module routine,
        and then runts its sanity_check, passing the specified
        logger."""
        logger.info('Export [config] CONFhwrf to the environment '
                    'variable $CONFhwrf.')
        os.environ['CONFhwrf'] = self.getstr('config','CONFhwrf')
        logger.info('Attempt to load hwrf_expt module.')
        import hwrf_expt
        logger.info('Attempt to initialize hwrf_expt module.')
        hwrf_expt.init_module()
        logger.info("Run the hwrf_expt module's own sanity_check routine.")
        hwrf_expt.sanity_check(logger)

    def guess_default_values(self):
        """Tries to set default values for some mandatory conf
        options.  The default values come from either other options or
        from environment variables.  If no suitable default can be
        found, execution will continue, but later jobs may fail.

        Config options are as follows.  If $VAR appears, that refers
        to ENV["VAR"]:

          config.cycle -- the cycle to run as a ten digit date (2014091418)
            Taken from cycle if present, otherwise $YMDH

          config.storm_num -- the storm number as a priority 1 to 5.
            Taken from $storm_num or uses the default of 1.

          config.stormlabel -- "storm" with the storm number appended
            (ie.: storm5 if storm_num=5).  

          dir.HOMEhwrf -- set to HOMEhwrf or $HOMEhwrf

          dir.WORKhwrf -- set to WORKhwrf or $WORKhwrf

          dir.syndat -- tcvitals directory.  Default: $COMINARCH

          config.input_catalog -- input catalog (conf section) name.
            Default: $INPUT_CATALOG or "hwrfdata"

          config.PARAFLAG -- NO if you are NCEP Central Operations
            (NCO), and YES otherwise.  This is used to turn on or off
            DBNet alerts and other NCO-specific options. 

        In addition, the following directories are added to the [dir]
        section:

          USHhwrf -- the location of ush scripts and the parent
            directory of the hwrf, pom and produtil packages
          FIXhwrf -- the location of the HWRF fix directory
          JOBhwrf -- the location of the HWRF jobs directory.  This is
            not needed by normal users.
          EXhwrf -- the location of the HWRF scripts directory
          PARMhwrf -- the location of the HWRF parm/ directory
          EXEChwrf -- the location of the HWRF exec/ directory
          KICKhwrf -- the location of the HWRF kick_scripts directory.  
            This is not needed by normal users.
          utilexec -- the location of the HWRF nwport/exec or
            /nwprod/exec directory
          utilscript -- the location of the HWRF nwport/ush or
            /nwprod/ush directory

        If set, these variables will be copied to the [config] section:
          EXPT -- optional: the experiment identifier, which must be
            alphanumeric, and can contain underscores.  Default: HWRF
          SUBEXPT -- optional: the subexperiment identifier, which
            must be alphanumeric, and can contain underscores.
            Default: set to value of EXPT
        """
        ENV=os.environ
        logger=self.log()
        PARAFLAG=( ENV.get('PARAFLAG','YES')!='NO' )

        def set_default(section,option,default,env1=None,env2=None):
            if not self.has_option(section,option):
                if env1 is not None and env1 in ENV:
                    self.set(section,option,ENV[env1])
                elif env2 is not None and env2 in ENV:
                    self.set(section,option,ENV[env2])
                elif default is not None:
                    self.set(section,option,str(default))
                else:
                    logger.error(
                        'Cannot find suitable default for [%s] option %s'%(
                            section,option))

        set_default('config','case_root','HISTORY','CASE_ROOT')
        set_default('config','EXPT','HWRF','EXPT')
        set_default('config','SUBEXPT','{EXPT}','SUBEXPT')
        set_default('dir','HOMEhwrf',None,'HOMEhwrf')
        set_default('dir','WORKhwrf',None,'WORKhwrf','DATA')
        set_default('config','datastore','{WORKhwrf}/hwrf_state.sqlite3')
        set_default('config','storm_num','1','storm_num')
        set_default('config','stormlabel','storm{storm_num}')
        set_default('config','input_catalog','hwrfdata','INPUT_CATALOG')
        set_default('dir','syndat',None,'COMINARCH')
        set_default('dir','com',None,'COMOUT')
        set_default('config','PARAFLAG','YES','PARAFLAG')
       
        if not self.has_option('config','cycle'):
            if 'YMDH' in ENV:
                self.cycle=ENV['YMDH']
        cycle=self.cycle

        if 'NWPROD' in ENV:
            NWPROD='NWPROD',ENV['NWPROD']
        elif 'envir' in ENV and os.path.exists('/nw'+ENV['envir']):
            NWPROD='/nw'+ENV['envir']
        else:
            NWPROD='{HOMEhwrf}/nwport'

        def dirset(evar,deff,parent='{HOMEhwrf}'):
            if evar in ENV:
                self._conf.set('dir',evar,ENV[evar])
            else:
                self._conf.set('dir',evar,parent+'/'+deff.lower())

        dirset('FIXhwrf','fix')
        dirset('USHhwrf','ush')
        dirset('EXhwrf','scripts')
        dirset('EXEChwrf','exec')
        dirset('JOBhwrf','jobs')
        dirset('PARMhwrf','parm')
        dirset('utilscript','ush',NWPROD)
        dirset('utilexec','util/exec',NWPROD)
        dirset('KICKhwrf','kick_scripts')

    def make_holdvars(self,part1='{PARMhwrf}/hwrf_holdvars.txt',part2=None):
        """Creates the storm*.holdvars.txt file needed by the old
        ksh-based scripts.  This is done for backward compatibility
        only.  The two arguments (part1 and part2) are two files to
        pass through self.strinterp and then into the holdvars file.
        Part 1 is mandatory, but part2 is optional.  It also fills in
        a few custom derived variables:

          cap_run_gsi -- capitalized version of [config] section run_gsi
          cap_run_relocation -- capitalized version of [config]
            section run_relocation
          holdvars_model -- "COUPLED" if [config] section run_ocean is
            true, and "ATMOS" if it is false."""
        assert(isinstance(part1,basestring))
        out=list()
        logger=self.log()
        gsi_flag=self.getbool('config','run_gsi')
        self.set('holdvars','cap_run_gsi',('YES' if gsi_flag else 'NO'))

        reloc_flag=self.getbool('config','run_relocation')
        self.set('holdvars','cap_run_relocation',
                 ('YES' if reloc_flag else 'NO'))

        ocean_flag=self.getbool('config','run_ocean')
        self.set('holdvars','holdvars_model',
                 ('COUPLED' if ocean_flag else 'ATMOS'))

        with open(self.strinterp('dir',part1),'rt') as f:
            for line in f: 
                out.append(self.strinterp('holdvars',line.rstrip()))
        if part2 is not None:
            with open(self.strinterp(part2),'rt') as f:
                for line in f: 
                    out.append(self.strinterp(line.rstrip()))
        return '\n'.join(out) + '\n'
