__all__=['PullGFSInput','HSIGrabber','PullInputError']

import dateutil, os, os.path, sys, logging, re
import produtil.run, produtil.cd, produtil.fileop
import hwrf.numerics

from produtil.fileop import isnonempty, deliver_file
from produtil.run import checkrun,bigexe,alias,runstr,run
from hwrf.numerics import to_datetime, to_datetime_rel,to_timedelta

#tag_rstprod=alias(bigexe('tag_rstprod'))


class PullInputError(Exception): pass

########################################################################

class HSIGrabber(object):
    def __init__(self,filename,hsi,tar,logger):
        self.filename=filename
        self.hsi=hsi
        self.tar=tar
        self._listing=None
        self.basename=None
        self.logger=logger
        if self.filename[-7:]=='.tar.gz':
            self.tarflags='z'
        elif self.filename[-8:]=='.tar.bz2':
            self.tarflags='j'
        else:
            self.tarflags=''
    def __enter__(self): 
        self.pull_archive()
        return self
    def __exit__(self,etype,evalue,etraceback):
        if evalue is None or isinstance(evalue,GeneratorExit):
            self.delete_archive()
        elif self.basename is not None:
            self.logger.warning('%s: not deleting archive due to exception.'%(self.basename,))
    def delete_archive(self):
        """Try to delete the local copy of the archive, logging errors."""
        basename=self.basename
        if basename is None: return
        try:
            self.logger.info('%s: delete local copy of archive file'%(basename,))
            os.unlink(self.basename)
        except EnvironmentError as e:
            self.logger.error('%s: cannot unlink: %s'%(self.basename,str(e)),exc_info=True)
    def pull_archive(self):
        filename=self.filename
        self.logger.info('%s: pull archive via hsi -P get'%(filename,))
        checkrun(self.hsi['-P','get',filename])
        basename=os.path.basename(filename)
        if not isnonempty(basename):
            raise PullInputError('%s: after hsi get, did not end up with %s'%(filename,basename))
        self.basename=basename
    def setlisting(self,listing): self._listing=listing
    def dellisting(self): self._listing=None
    def getlisting(self):
        if self._listing is not None: return self._listing
        basename=self.basename
        self.logger.info('%s: get archive listing'%(basename,))
        tar=self.tar
        if self.tarflags!='':
            tar=tar['-'+self.tarflags]
        listing=runstr(tar['--list','--verbose','--file',basename])
        self._listing=[line.rstrip() for line in listing.split('\n')]
        self.logger.info('%s: found %d files'%(basename,len(self.listing)))
        return self._listing
    listing=property(getlisting,setlisting,dellisting,
        """The GNU Tar long-format listing of the archive contents.""")
    def get(self,list_of_files):
        if self.basename is None:
            raise PullInputError('Tried to call get() before calling __enter__ or pull_archive')
        basename=self.basename
        if len(list_of_files)==0:
            self.logger.warning('%s: called "HSIGrabber.get" with an empty list.  De-archive no files as requested!'%(basename,))
            return
        self.logger.info('%s: tar -%sixpf ... %s ... (%d files)'
                         %(basename,self.tarflags,
                           ' '.join(list_of_files[0:3]),len(list_of_files)))
        checkrun(self.tar['-%sixpf'%(self.tarflags,),basename][list_of_files])
        for filename in list_of_files:
            if not os.path.exists(filename):
                raise PullInputError(
                    '%s: did not exist in %s (%s)'%
                    (filename,self.filename,self.basename))
        self.logger.info('%s: got %d files, all exist'%(basename,len(list_of_files)))
    def getall(self):
        if self.basename is None:
            raise PullInputError('Tried to call get() before calling __enter__ or pull_archive')
        self.logger.info('%s: tar -%sixpf ALL FILES!')
        checkrun(self.tar['-%sixpf'%(self.tarflags,),basename])

########################################################################

class PullGFSInput(object):
    def __init__(self,when,gfstarget,gdastarget,enkftarget,logger):
        self.hsi=alias(bigexe('hsi'))
        self.tar=alias(bigexe('tar'))
        self.when=to_datetime(when)
        self.prior=to_datetime_rel('-06:00',when)
        self.gfstarget=self.timename(gfstarget)
        self.enkftarget=self.priorname(enkftarget)
        self.gdastarget=self.priorname(gdastarget)
        self.logger=logger
    def log(self): return self.logger
    def timename(self,format):
        """Convenience function that uses self.when.strftime to generate one or more filenames."""
        if isinstance(format,list) or isinstance(format,tuple):
            return [ self.when.strftime(xx) for xx in format ]
        else:
            return self.when.strftime(format)

    def priorname(self,format):
        """Convenience function that uses self.prior.strftime to generate one or
more filenames based on the prior cycle's time."""
        if isinstance(format,list) or isinstance(format,tuple):
            return [ self.prior.strftime(xx) for xx in format ]
        else:
            return self.prior.strftime(format)

    def make_output_dirs(self):
        produtil.fileop.makedirs(self.gfstarget)
        produtil.fileop.makedirs(self.enkftarget)
        produtil.fileop.makedirs(self.gdastarget)

    def name_gfs_file(self,filename):
        """Generates the full path to the target location of the specified GFS file."""
        return os.path.join(self.gfstarget,filename)

    def name_enkf_file(self,filename):
        """Generates the full path to the target location of the specified ENKF file."""
        return os.path.join(self.enkftarget,filename)

    def name_gdas_file(self,filename):
        """Generates the full path to the target location of the specified GDAS file."""
        return os.path.join(self.gdastarget,filename)

    def process_enkf(self):
        filename=self.priorname('/NCEPPROD/hpssprod/runhistory/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_enkf.%Y%m%d_%H.fcs.tar')
        inputs = self.priorname(['./sfg_%Y%m%d%H_fhr06s_mem{i:03d}'.format(i=1+x) for x in xrange(80)])
        with HSIGrabber(filename,self.hsi,self.tar,self.logger) as arch:
            arch.get(inputs)
        for fn in inputs:
            deliver_file(fn,self.name_enkf_file(fn),keep=False,logger=self.log())

    def process_gdas(self):
        filename=self.priorname('/NCEPPROD/hpssprod/runhistory/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_gdas.%Y%m%d%H.tar')
        inputs = self.priorname(
            ['./gdas1.t%Hz.sf{h:02d}'.format(h=x+3) for x in xrange(7)] +
            ['./gdas1.t%Hz.pgrbh{h:02d}'.format(h=x+3) for x in xrange(7)] +
            ['./gdas1.t%Hz.abias', './gdas1.t%Hz.satang']  )
        with HSIGrabber(filename,self.hsi,self.tar,self.logger) as arch:
            arch.get(inputs)
        # Deliver a copy of abias and satang to current GFS cycle:
        for what in [ 'satang' ]:
            deliver_file(self.priorname('gdas1.t%Hz.'+what),
                         self.name_gfs_file(self.timename('gfs.t%Hz.'+what)),
                         keep=True,logger=self.log()) # note: keep=True here
        # Deliver everything to GDAS areas:
        for fn in inputs:
            deliver_file(fn,self.name_gdas_file(fn),keep=False,logger=self.log())

    def process_gfs_anl(self):
        filename=self.timename('/NCEPPROD/hpssprod/runhistory/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_gfs.%Y%m%d%H.anl.tar')
        with HSIGrabber(filename,self.hsi,self.tar,self.logger) as arch:
            keep=list()
            for line in arch.listing:
                if line.find('rstprod')>=0: continue # skip restricted file
                m=re.search('\s(./gfs\S*bufr_d|./avn.*atcfunix)$',line)
                if m: keep.append(m.group(1))
            if not keep:
                raise PullInputError('No unrestricted ./gfs*bufr_d files found in %s'%(arch.basename,))
            keep.extend(self.timename([
                        './gfs.t%Hz.sanl',
                        './gfs.t%Hz.sfcanl',
                        './gfs.t%Hz.abias',
                        './gfs.t%Hz.prepbufr.nr',
                        './gfs.t%Hz.syndata.tcvitals.tm00' ]))
            arch.get(keep)
        for subfile in keep:
            deliver_file(
                subfile,self.name_gfs_file(os.path.basename(subfile)),
                keep=False,logger=self.log())

    def process_gfs_gribs(self):
        grib2a=self.timename('/NCEPPROD/hpssprod/runhistory/1year/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_gfs.%Y%m%d%H.pgrb0p5.tar')
        grib2b=self.timename('/NCEPPROD/1year/hpssprod/runhistory/rh%Y/save/com_gfs_prod_gfs.%Y%m%d%H.pgrb0p5.tar')
        grib2=grib2a if(0==run(self.hsi['ls',grib2a])) else grib2b
        
        grib2files=self.timename(['./gfs.t%Hz.pgrb2f{x:02d}'.format(x=h*3) for h in xrange(44)])
        with HSIGrabber(grib2,self.hsi,self.tar,self.logger) as arch:
            arch.get(grib2files)
        for grib2file in grib2files:
            deliver_file(grib2file,
                self.name_gfs_file(grib2file),keep=False,logger=self.log())

        grib1a=self.timename('/NCEPPROD/hpssprod/runhistory/1year/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_gfs.%Y%m%d%H.pgrb.tar')
        grib1b=self.timename('/NCEPPROD/1year/hpssprod/runhistory/rh%Y/save/com_gfs_prod_gfs.%Y%m%d%H.pgrb.tar')
        grib1=grib2a if(0==run(self.hsi['ls',grib2a])) else grib2b

        grib1file=self.timename('./gfs.t%Hz.pgrbf00')
        with HSIGrabber(grib1,self.hsi,self.tar,self.logger) as arch:
            arch.get([grib1file])
        deliver_file(grib1file,self.name_gfs_file(grib1file),keep=False,logger=self.log())

    def process_gfs_sigma(self):
        filename=self.timename('/NCEPPROD/hpssprod/runhistory/2year/rh%Y/%Y%m/%Y%m%d/com_gfs_prod_gfs.%Y%m%d%H.sigma.tar')
        with HSIGrabber(filename,self.hsi,self.tar,self.logger) as arch:
            sigmas=[self.timename("./gfs.t%Hz.sf{x:02d}").format(x=x*3) for x in xrange(4)] \
                + [self.timename("./gfs.t%Hz.sf{x:02d}").format(x=x*6+12) for x in xrange(20)]
            arch.get(sigmas)
        for subfile in sigmas:
            deliver_file(subfile,self.name_gfs_file(subfile),keep=False,logger=self.log())

    def process_gdas_and_gfs_anl(self):
        self.make_output_dirs()
        self.process_gdas()
        self.process_gfs_anl()

    def process_all(self):
        self.make_output_dirs()
        self.process_enkf()
        self.process_gdas()
        self.process_gfs_anl()
        self.process_gfs_sigma()
        self.process_gfs_gribs()

