__all__=[] # Ensure that accidental "from hwrf_expt import *" does nothing

import os,os.path

import pdb
import produtil.datastore, produtil.run

import hwrf.config, hwrf.wrf, hwrf.post, hwrf.numerics, hwrf.launcher
import hwrf.regrib, hwrf.gribtask, hwrf.tracker, hwrf.storminfo
import hwrf.wps, hwrf.nhc_products, hwrf.copywrf, hwrf.fcsttask
import hwrf.relocate, hwrf.init, hwrf.prep, hwrf.gsi, hwrf.mpipomtc

from produtil.run import exe,alias
from hwrf.wrf import WRFDomain,WRFSimulation,ExternalWRFTask
from hwrf.post import PostManyWRF
from hwrf.regrib import RegribMany,igrb1,clatlon,GRIB2

def sanity_check(logger):
    """Runs a sanity check on this module's contents.  This should be
    called after init_module."""
    logger.info('The runwrf object = %s'%(repr(runwrf),))
    logger.info('The gribber object = %s'%(repr(gribber),))
    logger.info('The datastore (ds) object = %s'%(repr(ds),))
    logger.info('The wrf object = %s'%(repr(wrf),))

    # Make sure the mandatory configuration variables are specified:
    gsi_flag=conf.getbool('config','run_gsi') 
    ocean_flag=conf.getbool('config','run_ocean')
    reloc_flag=conf.getbool('config','run_relocation')
    spectral_flag=conf.getbool('config','use_spectral')
    fallbacks_flag=conf.getbool('config','allow_fallbacks')

    if ocean_flag:
        logger.info('The pominit object = %s'%(repr(pominit),))

    # Make sure the configuration makes sense and is supported:
    if gsi_flag and not reloc_flag:
        logger.error("Cannot use GSI without relocation.")
        raise hwrf.exceptions.HWRFConfigUnsupported(
            "Cannot use GSI without relocation.")

def init_module(CONFhwrf=None):
    global conf,ds,moad,storm1outer,storm1inner,wrf, anl4trak
    global runwrf,nonsatpost,satpost,hwrfsub,stormloc,domloc, gfs_init
    global gribber, tracker, WRFOUThwrf, coupled, pominit
    global nhcp, wrfcopier, real12, real126, WORKhwrf, HOMEhwrf
    global rel_stage1, rel_stage2, rel_stage3, fgat_init
    global gsi_d02, gsi_d03, gdas_merge, cycle

    # Figure out where to find the config file if it was not supplied
    # as an argument:
    if CONFhwrf is None:
        CONFhwrf=os.environ['CONFhwrf']
    
    # Generate the HWRFConfig object, which stores experiment
    # configuration, processor counts and so on.  It also knows how to
    # automatically generate other critical objects like the Datastore:
    conf=hwrf.launcher.load(CONFhwrf)
    conf.log().info('Initializing hwrf_expt module...')

    # Convenience variables:
    cycle=conf.cycle
    WORKhwrf=conf.getdir('WORKhwrf')
    HOMEhwrf=conf.getdir('HOMEhwrf')

    # Major configuration variables:
    satpost_flag=conf.getbool('config','run_satpost',True)
    gsi_flag=conf.getbool('config','run_gsi') 
    ocean_flag=conf.getbool('config','run_ocean')
    reloc_flag=conf.getbool('config','run_relocation')
    spectral_flag=conf.getbool('config','use_spectral')
    fallbacks_flag=conf.getbool('config','allow_fallbacks')

    # ----------------------------------------------------------------------
    # Define the HWRF workflow

    # Create and obtain the Datastore object, which stores the HWRF state
    # information:
    ds=conf.datastore

    # Create the objects in this module in a single transaction since that
    # speeds things up by a factor of 100:
    with ds.transaction():
        # Known types of domains:
        moad=WRFDomain(conf,'moad')
        storm1outer=WRFDomain(conf,'storm1outer')
        storm1inner=WRFDomain(conf,'storm1inner')
        storm1ghost_parent=WRFDomain(conf,'storm1ghost_parent')
        storm1ghost=WRFDomain(conf,'storm1ghost')
        storm1ghost_parent_big=WRFDomain(conf,'storm1ghost_parent_big')
        storm1ghost_big=WRFDomain(conf,'storm1ghost_big')
        postdoms=[moad,storm1outer,storm1inner]

        wrf=WRFSimulation(conf,'wrf',moad,conf.cycle,
                          conf.cycle+hwrf.numerics.to_timedelta(126*3600),45)
        wrf.add(storm1outer,moad)
        wrf.add(storm1inner,storm1outer)
        wrf.analysis_in(None)
        wrf.add_output('history',step=3600*3,end=9*3600)
        wrf.add_output('auxhist1',step=3600,outname='wrfdiag_d<domain>', \
                       frames_per_outfile=999,io_form=202)
        wrf.add_output('auxhist2',step=3600,end=9*3600,
                       outname='wrfout_d<domain>_<date>')
        wrf.add_output('auxhist3',step=3600*3,
                       outname='wrfout_d<domain>_<date>')
        #wrftask=ExternalWRFTask(ds,conf,'wrf',wrf,location=WRFOUThwrf)

        # Set up the wrfdoms.  This is the same as postdoms, but with
        # information filled in about grid IDs, output times, etc.:
        wrfdoms=[d for d in wrf]

        # ------------------------------------------------------------------
        # ATMOSPHERE PRE-PROCESSING

        wrfghost=WRFSimulation(conf,'wrf',moad,conf.cycle,
                            conf.cycle+hwrf.numerics.to_timedelta(6*3600),45)
        wrfghost.add(storm1ghost_parent,moad)
        wrfghost.add(storm1ghost,storm1ghost_parent)

        wrfghost_big=WRFSimulation(conf,'wrf',moad,conf.cycle,
                            conf.cycle+hwrf.numerics.to_timedelta(6*3600),45)
        wrfghost_big.add(storm1ghost_parent_big,moad)
        wrfghost_big.add(storm1ghost_big,storm1ghost_parent_big)

        gfs_init=hwrf.init.HWRFInit(
            ds,conf,'gfsinit',wrf,6*3600,126*3600,ibdystep=6*3600,
            wrfghost=wrfghost_big,prep=spectral_flag,track=True,
            realfcst=True,relocate=reloc_flag)

        if gsi_flag:
            fgat_init=hwrf.init.FGATInit(
                ds,conf,'fgat',wrf,6*3600,126*3600,wrfghost=wrfghost,
                prep=spectral_flag,track=True,realfcst=False,ibdystep=3*3600,
                in_dataset='gdas1',relocate=reloc_flag,gsi_d02=True,
                gsi_d03=True)
            ceninit=fgat_init.init_at_time(conf.cycle)

            ingsi_d02=fgat_init.get_relocates(storm1ghost_parent)
            gsi_d02=hwrf.gsi.FGATGSI(
                ds,conf,'gsi_d02',storm1ghost_parent,
                ceninit.rstage3.get_ghost(storm1ghost_parent),
                ingsi_d02,wrfghost)

            ingsi_d03=fgat_init.get_relocates(storm1ghost)
            gsi_d03=hwrf.gsi.FGATGSI(
                ds,conf,'gsi_d03',storm1ghost,
                ceninit.rstage3.get_ghost(storm1ghost),
                ingsi_d03,wrfghost)

            gdas_merge=hwrf.relocate.Merge(
                ds,conf,'merge',ceninit.rstage3,
                wrfinput=gfs_init.realinit,
                wrfanl=gfs_init.runwrfanl,
                taskname='gdas_merge',
                gsi_d02=gsi_d02, gsi_d03=gsi_d03)

        # ------------------------------------------------------------------
        # OCEAN PRE-PROCESSING
        if ocean_flag:
            pominit=hwrf.mpipomtc.POMInit(ds,conf,'pom')

        # ------------------------------------------------------------------
        # FORECAST 

        if ocean_flag:
            runwrf=hwrf.mpipomtc.WRFCoupledPOM(
                ds,conf,'runwrf',wrf,taskname='runwrf',pominit=pominit)
        else:
            runwrf=hwrf.fcsttask.WRFAtmos(
                ds,conf,'runwrf',wrf,taskname='runwrf')

        runwrf.add_metgrid(gfs_init.metgrid) \
              .add_geogrid(gfs_init.geogrid) \
              .add_fort65(gfs_init.realinit) \
              .add_wrfbdy(gfs_init.realfcst)

        # Add the merge output for the initial conditions and realfcst
        # for the boundary:
        if gsi_flag:
            runwrf.add_wrfinput(gdas_merge) \
                .add_wrfanl(gdas_merge,storm1outer) \
                .add_wrfanl(gdas_merge,storm1inner)

        if not gsi_flag or fallbacks_flag:
            if reloc_flag:
                runwrf.add_wrfinput(gfs_init.rstage3) \
                    .add_wrfanl(gfs_init.rstage3,storm1outer) \
                    .add_wrfanl(gfs_init.rstage3,storm1inner)
            else:
                # Do not fall back to "no relocation" even if
                # fallbacks are enabled.  This is because the
                # forecasting skill is useless without vortex
                # relocation.
                runwrf.add_wrfinput(gfs_init.realinit) \
                    .add_wrfanl(gfs_init.runwrfanl,storm1outer) \
                    .add_wrfanl(gfs_init.runwrfanl,storm1inner)

        # ------------------------------------------------------------------
        # POST-PROCESSING

        # Post-processors:
        nonsatpost = PostManyWRF(runwrf,postdoms,conf,'nonsatpost',3600,
                 needcrtm=False,streams=['history','auxhist2','auxhist3'])
        satpost    = PostManyWRF(runwrf,postdoms,conf,'satpost',6*3600,
                 needcrtm=True,streams=['auxhist3'])
        # WRF output copier:
        wrfcopier = hwrf.copywrf.WRFCopyTask(
            ds,conf,'copywrf',runwrf,conf.getstr('config','out_prefix'))
        wrfcopier.d_initial('namelist.input')
        wrfcopier.d_initial('wrfinput_d01')
        wrfcopier.d_initial('wrfbdy_d01')
        wrfcopier.d_initial('fort.65')
        wrfcopier.d_initial(wrf.analysis_name(storm1outer),
                            destname='{out_prefix}.wrfanl_d02')
        wrfcopier.d_initial(wrf.analysis_name(storm1inner),
                            destname='{out_prefix}.wrfanl_d03')
        for fhr in [0,3,6,9]: # deliver 0-9hr history stream to com
            for product in runwrf.products(time=3600*fhr,stream='history'):
                wrfcopier.d_wrfprod(product,check=False,
                  destname='{vit[stnum]:02d}{vit[basin1lc]}.{inname}')
        wrfcopier.d_final('track_d03.patcf')

        # ------------------------------------------------------------------
        # HWRF OUTPUT FILES
        
        # Domain selection.  These objects will grab Post output for the
        # specified domain for any arbitrary time when inserted into a
        # RegribMany object:
        grid3=igrb1(nonsatpost,domain=storm1inner)  # 3km non-satellite post
        grid2=igrb1(nonsatpost,domain=storm1outer)  # 9km non-satellite post
        grid1=igrb1(nonsatpost,domain=moad)         # 27km non-satellite post
        satE3=igrb1(satpost,domain=storm1inner)     # 3km satellite post
        satE2=igrb1(satpost,domain=storm1outer)     # 9km satellite post
        satE1=igrb1(satpost,domain=moad)            # 27km satellite post
        
        # The hwrfsub, when used in a RegribMany, selects the
        # hwrfprs_c and hwrfsat_c GRIB1/2 subset:
        hwrfsub=hwrf.regrib.GRIBSubsetter(
            'part',hwrf.tracker.hwrf_combine_subset,None)

        # In a RegribMany, trksub selects the subset needed by the tracker:
        trksub=hwrf.regrib.GRIBSubsetter(
            'part',hwrf.tracker.tracker_subset,None)

        # The domloc is the location of the HWRF domain center:
        domloc=hwrf.regrib.FixedLocation(lat=conf['config','domlat'],
                                         lon=conf['config','domlon'])
        basin=conf.syndat.pubbasin2
        if ((basin=='AL' or basin=='EP') and domloc.ewcenter<360.0):
            domloc.ewcenter+=360.0
       
        # Create the RegribMany object.  This object describes the GRIB
        # processing to do after the post completes:
        r=RegribMany(copygb=alias(exe(conf.getexe('copygb'))),
                     wgrib=alias(exe(conf.getexe('wgrib'))),
                     cnvgrib_g12=alias(exe(conf.getexe('cnvgrib'))
                                       ['-g12','-p32']))
       
        # Define some grids that we'll reuse later:
        r.add('d23',clatlon(grid2,res=[0.03,0.03],size=[12.,15.],
                            scan=136,n=[501,401]))
        r.add('d123low',clatlon(domloc,res=[0.25,0.25],size=[90.,110.],
                                scan=136,n=[441,361]))
        r.add('d123high',clatlon(domloc,res=[0.1,0.1],size=[90.,110.],
                                 scan=136,n=[1101,901]))
        r.add('d2',clatlon(grid2,res=[0.1,0.1],size=[12.,14.],
                           scan=136,n=[141,121]))
        r.add('d3',clatlon(grid3,res=[0.03,0.03],size=[7.5,9.0],
                           scan=136,n=[301,251]))
       
        r.add('hwrfprs_m',grid2*r.grid('d23')+grid3*r.grid('d23'))
        r.add('parenthigh',grid1/hwrfsub*r.grid('d123high'))
        # Need to add these four once they exist:
        r.add('trk',hwrf.tracker.expandlatlon \
                  (r.GRIB('hwrfprs_m'),west=2.5,east=2.5,north=4.0,south=4.0,
                  n=[667,667],scan=128,res=[0.03,0.03]))
        r.add('trkbasic',(r.GRIB('parenthigh')/trksub) \
                  .regridmerge(r.grid('trk'),r.GRIB('hwrfprs_m')/trksub))
        r.add('hwrftrk',hwrf.tracker.vinttave(r.GRIB('trkbasic')))
        
        r.add('hwrfprs_n',grid3*r.grid('d3'))
        r.add('hwrfprs_i',grid2*r.grid('d2'))
        r.add('hwrfprs_p',grid1*r.grid('d123low'))
        r.add('hwrfprs_c',grid1/hwrfsub*r.grid('d123high') +
              (grid2/hwrfsub*r.grid('d123high') +
               grid3/hwrfsub*r.grid('d123high')))
        
        if satpost_flag:
            r.add('hwrfsat_m',satE2*r.grid('d23')+satE3*r.grid('d23'))
            r.add('hwrfsat_n',satE3*r.grid('d3'))
            r.add('hwrfsat_i',satE2*r.grid('d2'))
            r.add('hwrfsat_p',satE1*r.grid('d123low'))
            r.add('hwrfsat_c',satE1/hwrfsub*r.grid('d123high') +
              satE2/hwrfsub*r.grid('d123high')) # sat_c only includes
                                                # d01 and d02
     
        # Add the deliveries to the RegribMany object:
        r.to_intercom('{out_prefix}.hwrftrk.grbf{fahr:02d}','hwrftrk')
        #r.to_intercom('{out_prefix}.hwrftrkbasic.grbf{famin:05d}m',
        #              'trkbasic')
        r.to_com('{out_prefix}.hwrftrk.grbf{fahr:02d}','hwrftrk')
        r.to_intercom('{out_prefix}.hwrfprs_m.grbf{fahr:02d}','hwrfprs_m')
        r.to_intercom('{out_prefix}.hwrfprs_n.grbf{fahr:02d}','hwrfprs_n')
        r.to_intercom('{out_prefix}.hwrfprs_i.grbf{fahr:02d}','hwrfprs_i')
        r.to_intercom('{out_prefix}.hwrfprs_p.grbf{fahr:02d}','hwrfprs_p')
        r.to_intercom('{out_prefix}.hwrfprs_c.grbf{fahr:02d}','hwrfprs_c')
        
	if satpost_flag:
            r.to_intercom('{out_prefix}.hwrfsat_m.grbf{fahr:02d}','hwrfsat_m')
            r.to_intercom('{out_prefix}.hwrfsat_n.grbf{fahr:02d}','hwrfsat_n')
            r.to_intercom('{out_prefix}.hwrfsat_i.grbf{fahr:02d}','hwrfsat_i')
            r.to_intercom('{out_prefix}.hwrfsat_p.grbf{fahr:02d}','hwrfsat_p')
            r.to_intercom('{out_prefix}.hwrfsat_c.grbf{fahr:02d}','hwrfsat_c')
        
        # Add GRIB2 converted versions of all GRIB1 products.  GRIB2,
        # unlike GRIB1, delivers the massive hwrfprs_m file.
        if satpost_flag:
            do_me=('prs','sat')
        else:
            do_me=('prs',)
        for T in do_me:
            for D in 'mnipc': # iterates over those five letters
                inx='hwrf%s_%s'%(T,D)    # looks like: hwrfprs_m
                outx='hwrf2%s_%s'%(T,D)  # looks like: hwrf2prs_m
                r.add(outx,r.GRIB(inx)*GRIB2)
                r.to_com('{out_prefix}.'+inx+'.grb2f{fahr:02d}',outx)
                    
        # Create the GRIBTask which will actually do the re-gribbing:
        gribber=hwrf.gribtask.GRIBTask(
            ds,conf,'regribber',r,start=conf.cycle,
            step=3600,end=126*3600)

        # Run the GFDL vortex tracker in multi-file moveable grid
        # mode off of the hwrftrk files from gribber:
        tracker=hwrf.tracker.TrackerTask(ds,conf,'tracker',
            start=conf.cycle, step=3600, end=126*3600)
        tracker.add_moving_grid(conf.syndat,gribber,'hwrftrk')
        tracker.send_raw_atcfunix('rawatcfunix',
            conf.strinterp('dir','{com}/{out_prefix}.trak.hwrf.raw'))
        track=tracker.send_atcfunix('cleanatcfunix',
            conf.strinterp('dir','{com}/{out_prefix}.trak.hwrf.atcfunix'))
        tracker.send_atcfunix('atcf3hourly',
            conf.strinterp('dir','{com}/{out_prefix}.trak.hwrf.3hourly'))
        tracker.send_atcfunix_subset('combinetrack',fhr=12,
            location=conf.strinterp(
                'dir','{com}/{vit[stnum]:02d}{basin1lc}.trak.hwrf.'
                'atcfunix.{YMDH}.combine'))

        # Run the hwrf_nhc_products program to generate the swath,
        # afos file and other custom outputs for NHC:
        nhcp=hwrf.nhc_products.NHCProducts(
            ds,conf,'nhc_products',runwrf,track,wrfdoms,conf.syndat)

    conf.log().info('Done in hwrf_expt module.')
