#! /usr/bin/env perl

use warnings;
use strict;
use File::Basename qw{dirname basename};
#use Time::HiRes qw {usleep};
use POSIX qw{WNOHANG WIFEXITED WEXITSTATUS ctime};
use Getopt::Std;

# WCOSS bug workaround:
sub usleep($) {
    select(undef,undef,undef,$_[0]/1e6)
}

my $UNSTARTED=0;
my $RUNNING=1;
my $DONE=2;
my $FAILED=3;

my @rsync='rsync';
my $verbose=0;
my $maxfail=3;
my $maxthreads=10;
my $username='hurdata';
my $hostname='jetscp.rdhpcs.noaa.gov';
#my $path="/pan2/projects/hwrf-vd/hwrf-input/FORECAST/gfs.2013080212/.";
my $path="/pan2/projects/hwrf-vd/hwrf-input/ENKF/2013/2013080212/.";
my %children;
my %fileinfo;
my %pidinfo;

sub verbose {
  if($verbose) {
    my $when=ctime(time());
    chomp($when);
    warn "$when: ",@_;
  }
}

sub warning {
  my $when=ctime(time());
  chomp($when);
  warn "$when: ",@_;
}

sub usage {
  die "ERROR: Format: $0 [options] user\@host:dirpath
 Options:
    -v    = verbose
    -t 10 = use up to 10 rsync threads (Default: $maxthreads)
    -f 5  = keep retrying up to 5 times (Default: $maxfail)

SCRIPT IS ABORTING DUE TO INCORRECT ARGUMENTS\n";
}

my %opts;
getopts("t:f:v",\%opts);
if($opts{v}) {
  $verbose=1;
  # Give a detailed stack trace on SIGINT:
  $SIG{INT} = sub {
    Carp::confess
  };
}

if($#ARGV<0) {
  usage();
}

$maxthreads=0+$opts{t} if defined($opts{t}) && $opts{t}>0;
$maxfail=0+$opts{f} if defined($opts{f}) && $opts{f}>0;

verbose "Settings:\n";
verbose "  max threads=$maxthreads\n";
verbose "  max tries=$maxfail\n";

sub nchildren {
  return scalar(keys(%children));
}

sub listchildren {
  return "{".join(", ",sort { $a <=> $b } keys %children)."}\n";
}

sub infanticide {
  # In UNIX, if a parent process does not kill its children, the
  # children become zombies.  This subroutine reaps child processes to
  # prevent that dire fate.

  # Whoever came up with UNIX terminology needs to see a therapist.

  my $pid=undef;
  my $status;
  while (1) {
    $pid=waitpid(-1 , WNOHANG);
    $status=0+$?;
    if (!defined($pid) || $pid<=0) {
      return;
    } elsif ($pid>0) {
      my $good=(WIFEXITED($status) && WEXITSTATUS($status)==0);
      my $fi=$pidinfo{$pid};
      if (defined($fi)) {
        if ($good) {
          $fi->{status}=$DONE;
          $fi->{exit_status}=$status;
          verbose "$fi->{name}: pid $pid completed normally ($status)\n";
        } else {
          $fi->{status}=$FAILED;
          $fi->{failcount}++;
          $fi->{exit_status}=$status;
          verbose "$fi->{name}: pid $pid failed ($status)\n";
        }
      } else {
        if($good) {
          verbose "process $pid: no file assigned (exited normally).\n";
        } else {
          verbose "process $pid: no file assigned (abnormal exit $status).\n";
        }
      }
      delete $children{$pid};
    }
    return if($pid==-1);
  }
}

sub background {
  # This launches an rsync command in the background and stores its
  # PID in %children.

  my $pid=fork();
  #    if(!(time()+0 < $starttime+120)) {
  #        die "TOO LATE";
  #    }
  if (defined($pid) && $pid>0) {
    $children{$pid}=1;
    return $pid;
  } elsif (defined($pid) && $pid==0) {
    # too verbose: verbose "EXEC: (".join(") (",@_).")\n";
    exec(@_);
    die "Cannot exec: ".join(" ",@_).": $?\n";
    exit(2);                    # should never get here
    while (1) {                 # Should also never get here
      sleep(1);
      die;
    }
  } else {
    die "Cannot fork: $?\n"
  }
}
;

sub get_file_list {
  warning "get_file_list: (rsync) (".join(") (",@_).")\n";
  open(RSYNC_LISTING,'-|','rsync',@_);
#  open(RSYNC_LISTING,"rsync ".join(" ",@_)." 2> /dev/null |");
    
  my @files;
    
  while (defined($_=<RSYNC_LISTING>)) {
    if (/^-r.*?(\S+)\s*$/) {
      push @files,$1;
    }
  }
  close(RSYNC_LISTING);
  return @files;
    
}

sub new_file {
  die "INTERNAL ERROR: wrong number of arguments to new_file" if $#_!=3;
  my ($filename,$remotepath,$localpath,$more)=@_;
  my $key="$localpath/$filename";
  $fileinfo{"$localpath/$filename"} = 
    {
     name=>"$filename",
     remotepath=>"$remotepath",
     localpath=>"$localpath",key=>$key,
     exit_status=>0,status=>$UNSTARTED,more=>$more,
     failcount=>0,pid=>0,failreport=>0
    };
  return $key;
}

sub start_transfer {
  my $filename=$_[0];
  #shift @_;
  my $fi=$fileinfo{$filename};
  my @more=@{$fi->{more}};
  my @verbose=();
  #push @verbose,'-v' if $verbose;
  my @command=(@rsync,"-a",@verbose,@more,
               "$fi->{remotepath}/$fi->{name}",
               dirname("$fi->{localpath}/$fi->{name}")."/.");
  
  $fi->{status}=$RUNNING;
  $fi->{command}=[@command];

  $fi->{pid}=background(@command);
  if (defined($fi->{pid}) && $fi->{pid}>0) {
    $pidinfo{$fi->{pid}}=$fi;
    return $fi->{pid};
  } else {
    # Could not start
    $fileinfo{$filename}{status}=$FAILED;
    return undef;
  }
}

sub add_rsync_list {
  my $source;
  my $dest='.';
  my @moreopt_in;
  if($#_==0) {
    verbose("One argument to add_rsync_list: {".join(", ",@_)."}\n");
    warning("$_[0]: invalid rsync source.  Must not begin with -\n") if $_[0]=~/^-/;
    $source=$_[0];
  } elsif($#_==1) {
    verbose("Two arguments to add_rsync_list: {".join(", ",@_)."}\n");
    warning("$_[0]: invalid rsync source.  Must not begin with -\n") if $_[0]=~/^-/;
    warning("$_[1]: invalid rsync destination.  Must not begin with - (use ./- instead)\n") if $_[1]=~/^-/;
    $source=$_[0];
    $dest=$_[1];
  } elsif($#_>1) {
    verbose("More than two arguments to add_rsync_list: {".join(", ",@_)."}\n");
    warning("$_[$#_-1]: invalid rsync source.  Must not begin with -\n") if $_[$#_-1]=~/^-/;
    warning("$_[$#_]: invalid rsync destination.  Must not begin with - (use ./- instead)\n") if $_[$#_]=~/^-/;
    @moreopt_in=@_[0..($#_-2)];
    $source=$_[$#_-1];
    $dest=$_[$#_];
  } else {
    # No arguments supplied
    warning("No arguments supplied to add_rsync_list.\n");
    return ();
  }

  if($source !~ /\A([^@]+)\@([^:]+):(.*)\z/) {
    warning "Internal error: invalid rsync source $source\n";
    return ();
  }
  my ($username,$hostname,$path)=($1,$2,$3);

  my $moreopt=[@moreopt_in];
  my $remotepath=$source;
  my $localpath=$dest;
  verbose "Listing rsync path $username\@$hostname:$path...\n";
  my @thesefiles=get_file_list("-arv",@$moreopt,"$username\@$hostname:$path/.");
  if ($#thesefiles<0) {
    warning "No files found.  Trying rsync protocol version 29...\n";
    @thesefiles=get_file_list("-arv","--protocol","29",@$moreopt,
                              "$username\@$hostname:$path/.");
    if ($#thesefiles>=0) {
      warning "  ...okay.  Protocol 29 worked.  Found ".(1+$#thesefiles)." file(s)\n";
      push @$moreopt,'--protocol','29'
    }
  } else {
    warning "Found ".(1+$#thesefiles)." file(s)\n";
  }

  my @out;
  my $file;
  foreach $file(@thesefiles) {
    push @out, new_file($file,$remotepath,$localpath,$moreopt);
  }
  return @out;
}

my @files;
my @args;

my $startarg=0;
for(my $i=0;$i<=$#ARGV;$i++) {
  if($ARGV[$i] eq ':') {
    my $stoparg=$i;
    if($stoparg>$startarg) {
      push @files,add_rsync_list(@ARGV[($startarg)..($stoparg-1)]);
    }
    $startarg=$i+1;
  }
}
do {
  my $stoparg=$#ARGV;
  if($stoparg>=$startarg) {
    push @files,add_rsync_list(@ARGV[($startarg)..($stoparg)]);
  }
};

my $donecount=0;
my $failcount=0;
my $giveupcount=0;
my $transferring=0;
my $filecount=$#files+1;
my $file;
while ($donecount+$giveupcount < $filecount) {
  foreach $file(@files) {
    infanticide();
    usleep(1e5);

    # Skip this file if we should skip it:
    if(defined($fileinfo{$file})) {
      # We have already started transferring this file.  Should we
      # start a new transfer?
      my %fi=%{$fileinfo{$file}};
      if($fi{status}==$DONE) {
        # too verbose: verbose "$file: will not re-transfer since we finished it already.\n";
        next;
      } elsif($fi{status}==$RUNNING) {
        # too verbose: verbose "$file: still transferring this one, so not re-transferring.\n";
        next;
      } elsif($fi{status}==$FAILED) {
        if($fi{failcount}>$maxfail) {
          verbose "$file: will not re-transfer since we gave up on it.\n";
          next;
        } else {
          verbose "$file: failed $fi{failcount} times but will retry.\n";
        }
      } else {
        # File is known, but has not yet been started.
      }
    }

    my $submitted=0;
    my $last_said_waiting=0;
    while (!$submitted) {
      infanticide();
      if (nchildren()<$maxthreads) {
        my $pid=start_transfer($file);
        verbose "$file: started process $pid\n";
        $submitted=1;
        infanticide();
        usleep(1e5);
        $last_said_waiting=0;
      } else {
        my $now=time();
        if($now>$last_said_waiting+10) {
          verbose "Waiting for processes.  Have ".nchildren()." running: ".listchildren();
          $last_said_waiting=$now;
        }
        infanticide();
        usleep(3e5);
      }
    }
  }

  # We have submitted transfers for every file.  Now check to see
  # how many are transferring, completed, etc.

  $donecount=0;
  $failcount=0;
  $giveupcount=0;
  $transferring=0;
  foreach $file(sort {$a cmp $b} keys %fileinfo) {
    my $stat=$fileinfo{$file}{status};
    if ($stat==$DONE) {
      $donecount++;
      # too verbose: verbose "  $file: done.\n"
    } elsif ($stat==$FAILED) {
      $failcount++;
      if ($fileinfo{$file}{failcount}>$maxfail) {
        if($fileinfo{$file}{failreport}<$fileinfo{$file}{failcount}) {
          warning "  $file: failed $fileinfo{$file}{failcount} times.  Giving up.\n";
          $fileinfo{$file}{failreport}=$fileinfo{$file}{failcount}
        }
        $giveupcount++
      } else {
        if($fileinfo{$file}{failreport}<$fileinfo{$file}{failcount}) {
          verbose "  $file: failed $fileinfo{$file}{failcount} times.  Will try again soon.\n";
          $fileinfo{$file}{failreport}=$fileinfo{$file}{failcount}
        }
      }
    } else {
      # too verbose: verbose "  $file: transfer already underway.\n";
      $transferring++;
    }
  }
  verbose "Status: $transferring transferring, $donecount done, $failcount failed (gave up on $giveupcount).\n";

  my $last_said_waiting=0;
  while(nchildren()>=$maxthreads) {
    infanticide();
    my $now=time();
    if($now>$last_said_waiting+10) {
      verbose "Waiting for processes.  Have ".nchildren()." running: ".listchildren();
      $last_said_waiting=$now;
    }
    infanticide();
    usleep(1e5);
  }
}

my $last_said_waiting=0;
while (nchildren()>0) {
  my $now=time();
  if($now>$last_said_waiting+10) {
    verbose "Waiting for processes.  Have ".nchildren()." running: ".listchildren();
    $last_said_waiting=$now;
  }
  infanticide();
  sleep 1;
}

my $notdone=0;
my $header=0;
foreach $file(@files) {
  my $done=1;
  if(!defined($fileinfo{$file})) {
    $done=0;
  } elsif($fileinfo{$file}{status}!=$DONE) {
    $done=0;
  }
  if(!$done) {
    if(!$header) {
      warning "Exiting.  Summary of failed files:";
      $header=1;
    }
    warning "$file: transfer failed.\n";
    $notdone++;
  }
}
if($notdone>0) {
  warning "Exiting with status 1 due to failed transfers.\n";
  exit(1);
} else {
  warning "Success! Exiting with status 0.\n";
  exit(0);
}
