#! /usr/bin/env perl

use warnings;
use strict;
use File::Basename;
use POSIX qw{fmod};

# Default table to parse if environment variable $RESERVATION_TABLE is unset:
my $default_reservation_table="/pan2/projects/hwrf-vd/$ENV{USER}/reservations.txt";

my $now_utc;  # hours since beginning of the day in UTC

# Special name for the hash key we use for the reservation chosen for
# this job.  The name used is impossible to find in the reservation
# table due to how the parser is written:
my $CHOSEN="== CHOSEN ==";

my $verbose=1;  # if true, then sub verbose() prints, otherwise it is silent

sub verbose {
  if($verbose) {
    warn @_;
  }
}

sub hhmm_to_hour($) {
  # Convert HH:MM or HH to hours (including fraction)
  my $in=$_[0];
  if($in=~/^0*(\d+)$/) {
    #                                   HH notation
    return 0+$1;
  } elsif($in=~/^0*?(\d+):0*?(\d+)$/) {
    #                                   HH:MM notation
    return( (0+$1) + (0.0+$2)/60.0 );
  }
}

sub now_utc() {
  # Returns the current time in UTC in hours since the beginning of
  # the day
  my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst)=gmtime();
  return $hour + ($min + ($sec+0.0) / 60.0) / 60.0;
}

sub match_date($$$) {
  # This date matching function checks to see if the job's date
  # ($ymdh) matches the storm/cycle line $linedate.  If the job's
  # month, day and hour are unknown then all dates match as long as
  # the year matches.
  my ($linedate,$ymdh,$full_ymdh_known)=@_;
  if($linedate=~/\A(\d\d\d\d)\z/) {
    # Only year specified
    return substr($ymdh,0,4)+0 == $1+0;
  } elsif($linedate eq '*') {
    return 1;
  } elsif($linedate=~/\A(\d\d\d\d)\%(\d\d)\z/) {
    # Only year and cycle specified
    if($full_ymdh_known) {
      return substr($ymdh,0,4)+0 == $1+0 && substr($ymdh,8,2)+0 == $2+0;
    } else {
      return substr($ymdh,0,4)+0 == $1+0;
    }
  } elsif($linedate=~/\A(\d{10})-(\d{10})\z/) {
    # Full date range specified
    my ($start,$end)=($1,$2);
    if($full_ymdh_known) {
      return $ymdh>=$start && $ymdh<=$end;
    } else {
      return substr($ymdh,0,4)>=substr($start,0,4) &&
        substr($ymdh,0,4)<=substr($end,0,4);
    }
  }
}

sub match_expt($$) {
  my ($lineexpt,$jobexpt)=@_;

  if (!defined($jobexpt) || $jobexpt eq '') {
    warn "    -- no expt specified for job (match by default)\n";
    return 1;
  } elsif (!defined($lineexpt) || $lineexpt eq '') {
    warn "    -- no expt specified on line (match by default)\n";
    return 1;
  }
  if($lineexpt eq '*') {
    # catch-all
    return 1;
  } elsif($lineexpt=~/^!(.*)$/) {
    # Negative match requested
    my $regex=$1;
    verbose "    -- $jobexpt !~ /$regex/\n";
    return !!($jobexpt !~ /$regex/);
  } else {
    # Positive match requested
    verbose "    -- $jobexpt =~ /$lineexpt/\n";
    return !!($jobexpt=~/$lineexpt/);
  }
}

sub match_kick($$) {
  my ($linekick,$jobkick)=@_;

  if (!defined($jobkick) || $jobkick eq '') {
    warn "    -- no kick specified for job (match by default)\n";
    return 1;
  } elsif (!defined($linekick) || $linekick eq '') {
    warn "    -- no kick specified on line (match by default)\n";
    return 1;
  }
  if($linekick=~/^!(.*)$/) {
    # Negative match requested
    my $regex=$1;
    verbose "    -- $jobkick !~ /$regex/\n";
    return !!($jobkick !~ /$regex/);
  } else {
    # Positive match requested
    verbose "    -- $jobkick =~ /$linekick/\n";
    return !!($jobkick=~/$linekick/);
  }
}

sub within_range($$$) {
  my ($now,$start_in,$duration_in)=@_;

  my $start=hhmm_to_hour($start_in);
  my $duration=hhmm_to_hour($duration_in);

  # Reservations with no start or duration (infinite reservations) always match:
  return 1 if(!defined($start) || !defined($duration));
  # If the time is unknown, do not match:
  return 0 if(!defined($now));

  verbose "start_in=$start_in becomes $start\n";
  verbose "duration_in=$duration_in becomes $duration\n";

  # Calculate the time since the start time.  If the start time is a
  # later hour, assume we mean yesterday.  That ensures that a 23:00
  # reservation of length 6:00 will still work at 4:30.
  my $diff=fmod($now-$start+240.0,24.0);

  # Allow the reservation if we're within the duration specified.
  if($diff<$duration+0.01) {
    return 1;
  } else {
    return 0;
  }
}

sub parse($$$$$$$) {
  # This is the parser for the reservation table file.  It handles one
  # line of the file after comments and blank lines are stripped, and
  # backslashed lines are concatinated.

  my $reslist=$_[0];
  my $line=$_[1];
  my $atcfid=$_[2];
  my $expt=$_[3];
  my $ymdh=$_[4];
  my $full_ymdh_known=$_[5];
  my $kick=$_[6];

  verbose "  input: reslist=$reslist line=$line\n    atcfid=$atcfid expt=$expt ymdh=$ymdh full=$full_ymdh_known kick=$kick\n";

  if($line=~/^\s*reservation\s+(\S+)(.*)$/i) {
    verbose "reservation block: $line\n";
    my ($resname,$rest)=($1,$2);
    if(defined($resname) && defined($rest) && 
       $resname ne '' && $rest ne '') {
      my (%res,$name,$val,$nvpair);
      my @nvpair=($rest=~/([^=[:blank:]]+=[^=[:blank:]]+)/g);
      foreach $nvpair (@nvpair) {
        ($name,$val)=($nvpair=~/(.*?)=(.*)/);
        $res{$name}=$val;
      }
      $res{name}=$resname;
      $reslist->{$resname}={%res};
      verbose "Added reservation $resname:\n";
      my $key;
      foreach $key (sort {$a cmp $b} keys %res) {
        verbose "  $key = \"$res{$key}\"\n";
      }
    }
  } elsif($line=~/^\s*(\d\d[a-zA-Z]|\*)\s+(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*(\S*)$/) {
    # This line specifies a storm and cycle list.  Does it match the
    # one for which this program is being called?
    my ($Xstid,$Xexpt,$Xres,$Xaccount,$Xdates,$Xkick)=($1,$2,$3,$4,$5,$6);
    verbose "  -- Match?\n";
    verbose "    line: stid=$Xstid expt=$Xexpt res=$Xres account=$Xaccount dates=$Xdates kick=$Xkick\n";
    verbose "    job: stid=$atcfid expt=$expt ymdh=$ymdh known=$full_ymdh_known kick=$kick\n";
    my $match_stid=("$Xstid" eq '*' || lc($Xstid) eq lc($atcfid));
    my $match_expt=match_expt($Xexpt,$expt);
    my $match_date=match_date($Xdates,$ymdh,$full_ymdh_known);
    my $match_kick=match_kick($Xkick,$kick);

    verbose "    stid match\n" if $match_stid;
    verbose "    stid mismatch\n" unless $match_stid;
    verbose "    expt match\n" if $match_expt;
    verbose "    expt mismatch\n" unless $match_expt;
    verbose "    date match\n" if $match_date;
    verbose "    date mismatch\n" unless $match_date;
    verbose "    kick match\n" if $match_kick;
    verbose "    kick mismatch\n" unless $match_kick;

    if($match_stid && $match_expt && $match_date && $match_kick) {
      # The line matches.  Is a reservation by this name defined in a
      # reservation block we already saw?
      verbose "  -- match\n";
      if(defined($reslist->{$Xres})) {
        # The reservation exists.  Are we within the allowed start time range?
        my $start=$reslist->{$Xres}{start};
        my $duration=$reslist->{$Xres}{duration};
        if(defined($start) && defined($duration)) {
          verbose "  -- check times start=$start duration=$duration\n";
        } else {
          verbose "  -- infinite reservation (no start or duration)\n";
        }
        if(!defined($start) || !defined($duration) ||
           within_range($now_utc,$start,$duration)) {
          # We can run in this reservation.
          verbose "  -- reservation block exists\n";
          if(!defined($reslist->{$Xres}{project})) {
            verbose "  -- missing project name, inserting the one from the storm/cycle line\n";
            $reslist->{$Xres}{project}=$Xaccount;
          }
          $reslist->{$CHOSEN}=$reslist->{$Xres};
        } else {
          verbose " -- not within reservation $Xres time window ($start+$duration)\n";
        }
      } else {
        # We have not seen this reservation, so define it based on the
        # information provided on this line.  This is just for
        # backward compatibility:
        verbose "  -- no reservation block\n";
        $reslist->{$CHOSEN}={ name=>"$Xres", project=>"$Xaccount" };
      }
    }
  }
}


my($atcfid,$expt,$dateinfo,$kickfull)=@ARGV;
my ($ymdh,$resfile);
my $full_ymdh_known=0;
my $kick='';

if(defined($kickfull) && $kickfull ne '') {
    $kick=basename($kickfull);
    $kick=~s/\..*//g;
}

if($#ARGV<2) {
  warn "ERROR: missing arguments.
SYNTAX: hwrf_reservation.pl 03L jet-2013-hwrf-parallel date-info [kick_script_name]
  where:
    03L -- storm id (case insensitive)
    jet-2013-hwrf-parallel -- HWRF \$EXPT or \$SUBEXPT variable
    date-info -- a year (2013) or year and hour (2013%18) or full date
       range (2013060100-2013060800)
    kick_script_name -- Optional: the basename of the
       kick_script/*.shell file being considered.
ABORTING due to missing arguments";
  exit 1;
}

verbose "kick = $kick\n";

if($dateinfo=~/\A\d\d\d\d\z/) {
  # When only the year is specified, fill in zeros for month, day and
  # hour:
  $ymdh=sprintf("%04d000000",0+$dateinfo);
  $full_ymdh_known=0;
} elsif($dateinfo=~/\A\d{10}\z/) {
  $ymdh="$dateinfo";
  $full_ymdh_known=1;
} else {
  die "ABORTING: Invalid date information $dateinfo.  It must be either a year (2013) or a ten digit YYYYMMDDHH string (ie.: July 4, 2013 at noon UTC would be 2013070412)\n"
}

if(defined($ENV{RESERVATION_TABLE})) {
  $resfile=$ENV{RESERVATION_TABLE};
} else {
  $resfile=$default_reservation_table;
}

# Get the local time:
$now_utc=now_utc();

# Read the entire reservation file into an array @res:
if(!open(RES,"<$resfile")) {
  # No reservation table file.  This is not an error.
  warn "Cannot open $resfile\n  errno says: $?\n  aborting";
  exit 0;
}
my @res=<RES>;
close RES;
chomp @res;

my %reslist;
my $line='';
foreach $_ (@res) {
  s/^#.*$//g; # strip comments that begin with a #
  /^ *$/ and next; # ignore blank lines
  /^\\$/ and next; # ignore lines that are only a backslash
  if(/^((?:[^\\]*|\\\\|\\[^\\])*)\\$/) {
    # This is a line ending in a \ so it is continued to the next line.
    $line.=$1;
    next;
  } else {
    $line.=$_;
  }
  verbose "PARSE $line\n";
  verbose "  supply atcfid=$atcfid expt=$expt ymdh=$ymdh full=$full_ymdh_known kick=$kick\n";
  parse(\%reslist,$line,$atcfid,$expt,$ymdh,$full_ymdh_known,$kick);
  $line='';
}
if($line ne '') {
  verbose "PARSE $line\n";
  verbose "  supply atcfid=$atcfid expt=$expt ymdh=$ymdh full=$full_ymdh_known kick=$kick\n";
  parse(\%reslist,$line,$atcfid,$expt,$ymdh,$full_ymdh_known,$kick);
  $line='';
}

if(defined($reslist{$CHOSEN})) {
  print qq(export JET_RESERVATION="$reslist{$CHOSEN}{name}");
  print qq( ACCOUNTLL="$reslist{$CHOSEN}{project}");

  if(defined($reslist{$CHOSEN}{partition})) {
    my $var;
    my $part=$reslist{$CHOSEN}{partition};
    foreach $var(qw{GROUPLL PARCLASSLL SERCLASSLL BIGCLASSLL
                    FORECASTCLASSLL ARCHIVECLASSLL TRANSFERCLASSLL
                    RUNDIRCLASSLL}) {
      if(defined($ENV{$var}) && ! -z $ENV{$var} && $ENV{$var} ne 'service') {
        print qq( $var="$part");
      }
    }

    print " JET_NAME=".substr($part,0,4);

    if($part eq 'tjet' || $part eq 'ujet' || $part eq 'tjet:ujet'
       || $part eq 'ujet:tjet') {
      print " WHICH_JET=t";
    } elsif($part eq 'njet') {
      print " WHICH_JET=n";
    } elsif($part eq 'sjet') {
      print " WHICH_JET=s";
    } else {
      # Assume tJet/uJet and hope for the best
      print " WHICH_JET=any";
    }
  }

  print "\n";
} else {
  # Could not find a reservation.  This is not an error.
  warn "Cannot find a reservation for this job.\n";
  exit 0;
}
