#! /usr/bin/perl -w

use Carp;
use strict;

my @input_files;

sub usage {
    die qq{
You are not calling $0 correctly.
Format: $0 [var=value [var2=value2 [...]]] INPUTFILE OUTPUTFILE

  INPUTFILE -- a readable input file

  OUTPUTFILE -- either a dash ("-") indicating stdout OR the full path to
      an output file.

  [var=value [var2=value2 [...]]] -- a list of variables to set.  These
      variables will be treated as if they had been set as environment
      variables before this script was run.

  This script will read in the contents of the input file, and replace any
  text of the form @[VARNAME] with the respective environment variable.

@_
}
}

sub die_handler {
    # An error occurred.  Send the error message to both stdout and stderr:
    warn @_;

    # Run llcancel to terminate the entire job step:
    if(defined($ENV{LOADL_STEP_ID})) {
        warn "ABORTING JOB VIA LLCANCEL $ENV{LOADL_STEP_ID}\n";
        system('llcancel',$ENV{LOADL_STEP_ID});
        sleep(5);
        warn "Called llcancel.  Job step hasn't aborted yet... ???\n";
    } elsif(defined($ENV{PBS_JOBID})) {
        warn "ABORTING JOB VIA QDEL $ENV{PBS_JOBID}\n";
        system('qdel',$ENV{PBS_JOBID});
        sleep(5);
        warn "Called qdel.  Job step hasn't aborted yet... ???\n";
    } elsif(defined($ENV{JOB_ID})) {
        warn "ABORTING JOB VIA QDEL $ENV{JOB_ID}\n";
        system('qdel',$ENV{JOB_ID});
        sleep(5);
        warn "Called qdel.  Job step hasn't aborted yet... ???\n";
    } elsif(defined($ENV{LSB_JOBID})) {
        warn "ABORTING JOB VIA BKILL $ENV{LSB_JOBID}\n";
        system('bkill',$ENV{LSB_JOBID});
        sleep(5);
    } else {
        warn "ERROR: Environment variables JOB_ID, PBS_JOBID, LSB_JOBID and LOADL_STEP_ID are not defined.  I cannot abort via qdel, bkill, or llcancel!  Maybe this process is not part of a SGE, Torque or LoadLeveler job?  Darn.  I guess I'll just exit with status 2 instead.\n";
    }

    # Exit the script with an error message (sent to stdout & stderr)
    exit(2);
}

my %functions=(
    'lc'=>sub { lc($_[0]) },
    'uc'=>sub { uc($_[0]) },
    'len'=>sub { length($_[0]) },
    'trim'=>sub {
        my $val=shift;
        $val=~/\A\s*(.*?)\s*\z/ms;
        $val=$1;
        $val='' unless defined($val);
        return $val;
    }
);

sub apply {
    my $val=shift;
    my $fun1=shift;
    my $morefun=shift;
    my $runme=$functions{$fun1};

    if(defined($runme)) {
        $val=$runme->($val);
        $val='' unless defined($val);
    } else {
        warn "Ignoring unknown function \"$fun1\" -- I only know these: ".join(', ',keys %functions)."\n";
    }
    if($morefun=~/\A\.([A-Za-z0-9_]+)(.*)\z/) {
        return apply($val,$1,$2);
    }
    return $val;
}

sub require_var {
    my $varname=shift();
    if($varname=~/\A([A-Za-z0-9_]+)\.([A-Za-z0-9_]+)(.*)\z/) {
        # This is a variable with requested function calls
        my ($fun1,$morefun);
        ($varname,$fun1,$morefun)=($1,$2,$3);
        my $val=require_var($varname);
        return apply($val,"$fun1","$morefun");
    } else {
        my $val=$ENV{$varname};
        if(!defined($val) ) {
            die "Error in input file \"$input_files[$#input_files]\": the \"$varname\" environment variable is not defined.  Results of call to env:\n".`env`."\nSCRIPT IS ABORTING BECAUSE \$$varname ENVIRONMENT VARIABLE IS NOT SET.\n";
        }
        return $val;
    }
}

sub optional_var {
    my $varname=shift();
    if($varname=~/\A([A-Za-z0-9_]+)\.([A-Za-z0-9_]+)(.*)\z/) {
        # This is a variable with requested function calls
        my ($fun1,$morefun);
        ($varname,$fun1,$morefun)=($1,$2,$3);
        my $val=optional_var($varname);
        return apply($val,"$fun1","$morefun");
    } else {
        my $val=$ENV{$varname};
        if(!defined($val) ) {
          return "";
        }
        return $val;
    }
}

sub replace_backslashed( $ ) {
    my $out;
    if($_[0]=~/\A[0-9]/) {
        $out= chr(oct($_[0]));
    } else {
        $out=$_[0];
    }
    return $out;
}

sub replace_vars( $ ) {
    my $text=shift;
    die unless defined($text);
    $text=~ s/(?<!\\)\$([a-zA-Z_0-9.]+)/${\require_var($1)}/g;
    $text=~ s/(?<!\\)\$\{([^{}]*)\}/${\var_or_command($1)}/g;
    $text=~ s/\\n/\n/g;
    $text=~ s/\\t/\t/g;
    $text=~ s/\\r/\r/g;
    $text=~ s/\\b/\b/g;
    $text=~ s/\\([0-9]{3}|.)/${\replace_backslashed("$1")}/g;
    return $text;
}

sub require_file( $ ) {
    my $filename=replace_vars(shift());
    open(EMBEDDED_FILE,"<$filename")
        or die "Unable to open file \"$filename\" for reading: $!";
    my $data;
    do {
        local $/;
        $data=<EMBEDDED_FILE>;
    };
    if(!defined($data)) {
        die "Error reading from file \"$filename\": $!";
    }
    close(EMBEDDED_FILE) or die "Error closing input file \"$filename\": $!";
    return $data;
}

sub var_or_command( $ ) {
    my $data=shift;
    $data=~/\A([a-z_A-Z][a-zA-Z_0-9]*)((?:\.[A-Za-z0-9.]+)?)(?:(==|!=|:\+|:-|=|:=|:\?|<|:<|:)(.*))?\z/ms;
    my ($varname,$functions,$operator,$operand)=($1,$2,$3,$4);
    $functions='' unless defined($functions);
    #warn "varname($varname) operator($operator) operand($operand)";
    if(defined($operator) && $operator ne '') {
        $operand='' unless defined($operand);

        my $vartext=$ENV{$varname};
        my $varset=(defined($ENV{$varname}) && $ENV{$varname} ne '');
        if($functions ne '') {
            $vartext='' unless defined($vartext);
            $functions=~/\A\.([A-Z0-9a-z_]+)(.*)\z/;
            my ($fun,$morefun)=($1,$2);
            $vartext=apply($vartext,$fun,$morefun);
        }

        if($operator eq ':+') {
            return $varset ? replace_vars($operand) : '';            
        } elsif($operator eq ':-') {
            return $varset ? $vartext : replace_vars($operand);
        } elsif($operator eq ':') {
            my $val=$vartext;
            $val='' unless defined $val;
            my ($start,$count) = ($operand=~/\A([0-9]+)(?::([0-9]+))?/ms);
            #warn "substr start($start) count($count) of val($val)";
            my $len=length($val);
            $start=0 unless defined($start);
            return '' if($start>=$len);
            $count=length($val)-$start unless defined($count);
            if($start+$count>$len) {
                $count=$len-$start;
            }
            return substr($val,$start,$count);
        } elsif($operator eq '=') {
            $ENV{$varname}=replace_vars($operand);
            return '';
        } elsif($operator eq '==' || $operator eq "!=") {
            # This is the ternary ?: operator.
            my $val=$vartext;
            my ($test,$then,$else) = ($operand=~/\A((?:[^\\\?]|(?:\\\\)*|(?:\\\\)*\\.)*)\?(.*?):((?:[^\\:]|(?:\\\\)*|(?:\\\\)*\\.)*)\z/ms);
            #warn "then($val)==($test) then($then) else($else)";
            $val="" unless defined $val;
            $test="" unless defined($test);
            $test=replace_vars($test);
            $then="" unless defined($then);
            $else="" unless defined($else);
            if($operator eq "==") {
                return replace_vars(  ($val eq $test) ? $then : $else );
            } else {
                return replace_vars(  ($val ne $test) ? $then : $else );
            }
        } elsif($operator eq ':=') {
            $ENV{$varname}=replace_vars($operand) unless $varset;
            return $ENV{$varname};
        } elsif($operator eq ':?') {
            return $vartext if $varset;
            $operand=replace_vars($operand);
            die "$varname: you did not define this variable.  Aborting" if($operand eq '');
            die "$varname: $operand";
        }
    } elsif(defined($varname)) {

        #warn "call require_var($varname$functions)";
        return require_var("$varname$functions");
    }
    die "Don't know what to do with text \"$data\"";
}

sub require_data( $ ) {
    my $data=shift;
    if($data=~/\A<(.*)\z/ms) {
        # This is an instruction to read in a file.
        return require_file($1);
    } elsif($data eq '@') {
        return '@'; # @[@] is replaced with an @
    } elsif($data =~ /\A\#/) {
        if($data=~/\@\[/) {
            die "Found a \@\[ construct nested within a comment (\@\[#...\]).  Aborting";
        }
        return ''; # @[#stuff] is a comment
    } else {
        # This is a variable name, a command or an error.
        return var_or_command($data);
    }
}

sub make_filename( $$ ) {
    my ($dirname,$filename)=@_;
    my $out=$dirname.'/'.$filename;
    $out=~ s://+:/:g; # Remove duplicate /'s
    return $out;
}

# Set up our special error handler which will run llcancel or qdel if
# any errors occur:
$SIG{__DIE__}=\&die_handler;

while($#ARGV>1) {
  $_=shift @ARGV;
  if( /\A([A-Z_a-z][A-Z_a-z0-9.]*)=(.*)\z/ ) {
    $ENV{$1}=$2;
  } else {
    usage "Unable to parse parameter \"$_\".  Aborting";
  }
}

if($#ARGV!=1) {
    die q{
You are not calling $0 correctly.
Format: $0 INPUTFILE OUTPUTFILE

  INPUTFILE -- a readable input file

  OUTPUTFILE -- either a dash ("-") indicating stdout OR the full path to
      an output file.

  This script will read in the contents of the input file, and replace any
  text of the form @[VARNAME] with the respective environment variable.
}
}

my ($infile,$outfile)=@ARGV;

push @input_files,$infile;

my $filecontents;

if($infile ne '-') {
    open(INFILE,"< $infile")
        or die "Unable to open input file \"$infile\" for reading: $!";
}
if($outfile ne '-') {
    open(OUTFILE,"> $outfile")
        or die "Unable to open output file \"$outfile\" for writing: $!";
}

# Parser state stack:
my @states=();

# Parser states.  Meanings of the hash entries:
#
#   active -- true = the parser should generate text for the current block
#
#   in_if_block -- true = we're inside an if/elseif/else/endif block,
#       but have not yet seen an "else"
#
#   in_ifelse_block -- true = we have passed the "else" in an
#       if/elseif/else/endif block
#
#   used_if -- true = we're in an if/elseif/else/endif block and we
#       have already used one of the ifs or elseifs
#
#   ignore -- we are in an if/elseif/else/endif block, but we are
#       ignoring the whole block because it is inside another,
#       inactive block (like an if or elseif that did not match).
my $outer={active=>1,in_if_block=>0,in_ifelse_block=>0,used_if=>0,ignore=>0};
my $if_unused_if={active=>0,in_if_block=>1,in_ifelse_block=>0,used_if=>0,ignore=>0};
my $if_active_if={active=>1,in_if_block=>1,in_ifelse_block=>0,used_if=>1,ignore=>0};
my $if_used_if={active=>0,in_if_block=>1,in_ifelse_block=>1,used_if=>1,ignore=>0};
my $if_active_else={active=>1,in_if_block=>0,in_ifelse_block=>1,used_if=>1,ignore=>0};
my $if_inactive_else={active=>0,in_if_block=>0,in_ifelse_block=>1,used_if=>1,ignore=>0};

my $ignore_if_block={active=>0,in_if_block=>1,in_ifelse_block=>0,used_if=>0,ignore=>1};
my $ignore_else_block={active=>0,in_if_block=>0,in_ifelse_block=>1,used_if=>0,ignore=>1};

my $line=0;

# Initialize the parser stack:
@states=($outer);

sub print_state {
    my $state;
    print STDERR "STATE STACK: \n";
    foreach $state (@states) {
        print STDERR "state: ";
        if($state->{ignore}) {
            print STDERR 'ignoring block: ';
        }
        print( STDERR $state->{active} ? 'active ' : 'inactive ' );
        if($state->{in_if_block}) {
            print STDERR 'in if block, before else ';
        }
        if($state->{in_ifelse_block}) {
            print STDERR 'in if block, after else ';
        }
        if(!$state->{in_if_block} && !$state->{in_ifelse_block}) {
            print STDERR 'not if or else';
        }
        if($state->{used_if}) {
            print STDERR '(have activated a past if/elseif/else) ';
        }
        print STDERR "\n";
    }
    print STDERR "END\n";
}

sub active {
    return 1 if($#states<0);
    my $x;
    foreach $x (@states) {
        return unless $x->{active};
    }
    return 1;
}
sub top_state { 
    if(defined($_[0])) {
        if(!defined($states[$#states])) {
            die "Internal error: no state to search when looking for $_[0] in top state";
        } elsif(!defined($states[$#states]{$_[0]})) {
            die "Internal error: cannot find $_[0]";
        }
        return $states[$#states]{$_[0]}
    } else {
        return $states[$#states]
    }
}
sub push_state {
    push @states,$_[0]
}
sub pop_state {
    pop @states;
}
sub replace_state {
    $states[$#states]=$_[0]
}

my $out_of_block=-1;
my $if_unused=1;
my $if_in_if=1;

sub parse_line {
    $line++;
#    print_state();
#    print STDERR "LINE: $_\n";

    if($_=~/^\s*\@\*\*\s*if\s+([A-Za-z_][A-Za-z_0-9.]*)\s*([!=])=\s*(.*?)\s*$/) {
        # This is the beginning of an IF block
        if(!active()) {
            # This IF lies within an inactive block, so we skip this
            # whole if, elseif, else, endif block.
            push_state($ignore_if_block);
            return;
        }
        my ($left,$cmp,$right)=($1,$2,$3);
        $left=optional_var($left);
        $right=replace_vars($right);
        if($left eq $right) {
            push_state(($cmp eq '=') ? $if_active_if : $if_unused_if);
        } else {
            push_state(($cmp eq '=') ? $if_unused_if : $if_active_if);
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*abort\s+(.*)$/) {
        if(active()) {
            die "Found an abort directive on line $line: $1\n";
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*warn\s+(.*)$/) {
        if(active()) {
            warn replace_vars($1)."\n";
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*else\s*if\s+([A-Za-z_][A-Za-z_0-9.]*)\s*([!=])=\s*(.*?)\s*\z/) {
        return if top_state('ignore');
        my ($left,$cmp,$right)=($1,$2,$3);
        $left=optional_var($left);
        $right=replace_vars($right);
        if($#states<0) {
            die "Found an elseif without matching if at line $line";
        }
        if(!top_state('in_if_block')) {
            if(top_state('in_ifelse_block')) {
                die "Unexpected elseif after an else at line $line";
            } else {
                die "Unexpected elseif at line $line";
            }
        } elsif(top_state('used_if')) {
            # the "if" or a prior elseif matched, so we ignore this
            # elseif and deactivate the block so all future
            # if/else/elseif will be unused.
            replace_state($if_used_if);
        } elsif(!top_state('active')) {
            my $activate;
            if($left eq $right) {
                $activate=(($cmp eq '=') ? 3 : 0);
            } else {
                $activate=(($cmp eq '=') ? 0 : 3);
            }
            replace_state($if_active_if) if $activate;
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*else\s*(?:\#.*)?$/) {
        if(top_state("used_if")) {
            # a prior if or elseif matched so we do not use this else
            replace_state($if_inactive_else);
        } elsif(top_state('in_ifelse_block')) {
            die "Found an extra else at line $line";
        } elsif(!top_state('in_if_block')) {
            die "Found an else outside an if at line $line";
        } elsif(top_state('ignore')) {
            # We're ignoring a whole if/elseif/else/endif block
            # because it lies within an inactive block.
            replace_state($ignore_else_block);
        } elsif(!top_state('used_if')) {
            replace_state($if_active_else);
        } else {
            # a prior if or elseif matched, so this block is now inactive
            replace_state($if_inactive_else);
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*endif\s*(?:\#.*)?$/) {
        if(top_state('in_if_block') || top_state('in_ifelse_block')) {
            pop_state();
        } else {
            die "Found an endif without matching if at line $line";
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*insert\s*(\S.*?)\s*$/) {
        if(active()) {
            my $contents=require_file($1);
            if($outfile eq '-') {
                print $contents;
            } else {
                print OUTFILE $contents;
            }
        }
        return;
    } elsif($_=~/^\s*\@\*\*\s*include\s*(\S.*?)\s*$/) {
        if(active()) {
            my @contents=split(/[\r\n]/,require_file($1));
            my $templine=$line;
            my $data;
            $line=0;
            foreach $data(@contents) {
                $_="$data\n";
                parse_line("$data");
            }
            $line=$templine;
            return;
        } else {
            return;
        }
    } elsif($_=~/^\s*\@\*\*.*/) {
        my $chomped=$_;
        chomp $chomped;
        warn "Invalid \@** directive in line \"$chomped\".  Ignoring line.\n";
        return;
    } elsif($#states>=0) {
        if(!active()) {
            return;
        }
    }
    
    # Replace text of the form @[VARNAME] with the contents of the
    # respective environment variable:
    s/\@\[((?:\n|[^\]])*)\]/${\require_data($1)}/msg;
    if($outfile eq '-') {
        print $_;
    } else {
        print OUTFILE $_;
    }
    if($line>1000000) {
        # Abort at 1,000,000 lines.
        die "Read more than a million lines from the input file.  Something is probably wrong";
    }
}

if($infile ne '-') {
    parse_line() while(<INFILE>);
} else {
    parse_line() while(<STDIN>);
}

if($#states!=0) {
    die "Not enough endifs";
}

if($outfile ne '-') {
    close(OUTFILE)
        or die "Error closing output file \"$outfile\": $!";
}
if($infile ne '-') {
    close(INFILE)
        or die "Error closing input file \"$infile\": $!";
}

