"""This module is intended to be used to communicate with the batch
system.  At present, it just knows how to guess the job name and id,
as well as a "longname" that combines the two."""

__all__=['jobname','jobid','joblongname','NONAME']
import os

class FakeClass: pass
UNSPECIFIED=FakeClass()
NONAME="NONAME"

def getenvs(names,fallback=None):
    """Tries the list of environment variable names, returning the first one
that exists and is non-blank.  If none are found, returns the
fallback."""
    for name in names:
        if name not in os.environ:           continue
        val=os.environ[name]
        if not isinstance(val,basestring):   continue
        if len(val)<1:                       continue
        return val
    return fallback

def jobname(fallback=UNSPECIFIED):
    """Returns the human-readable job name, if one exists.  Attempts to get
it from the NCO $outid or $job environment variables first, then tries
the batch system variables.  If none are found, and fallback is
specified, then the fallback is returned.  Otherwise, the module-level
NONAME variable is returned (which defaults to "NONAME")."""
    ret=getenvs(['outid','job','LSB_JOBNAME','PBS_JOBNAME','MOAB_JOBNAME',
                'LOADL_STEP_NAME','LOADL_JOB_NAME'])
    if ret is not None: return ret
    if fallback is not UNSPECIFIED: return fallback
    return NONAME

def jobid(fallback=UNSPECIFIED):
    """Returns the batch system job id for the batch job that is running this
program, if known.  Tries the NCO $pid first, then the various batch
system environment variables.  If none are found, and the fallback is
specified, returns the fallback.  Otherwise, returns "o$PID" where
$PID is the process ID."""
    ret=getenvs(['pid','LSB_JOBID','PBS_JOBID','MOAB_JOBID','LOADL_STEP_ID'])
    if ret is not None: return None
    if fallback is not UNSPECIFIED: return fallback
    return 'o%s'%(str(os.getpid()),)

def joblongname(jobid_fallback=UNSPECIFIED,
                jobname_fallback=UNSPECIFIED):
    """Returns a human-readable job name that includes both the batch system
job name and id.  Returns the NCO $jobid variable if available,
otherwise returns LL{jobid()}.o{jobname()} where jobid and jobname are
the results of those two functions.  The jobid_fallback and
jobname_fallback are passed as the fallback parameters to the calls to
jobid and jobname"""
    ret=getenvs(['jobid'])
    if ret is not None: return ret
    return 'LL%s.%s'%(jobid(jobid_fallback),jobname(jobname_fallback))
