"""This module configures logging for stdout, stderr and the jlogfile.
It also contains the jlogger, a logger.Logger object that is used to
log directly to the jlogfile, and jlogdomain: a string name of the 
logger domain for the jlogfile."""

__all__ = [ 'configureLogging','jlogger','jlogdomain','postmsg',
            'MasterLogFormatter','JLogFormatter','stdout_is_stderr',
            'MasterLogHandler','JLogHandler' ]

import logging, os, sys, traceback
import produtil.batchsystem

logthread=''  # string for log messages to indicate thread number/name 
jlogdomain='jlog'
jlogger=logging.getLogger(jlogdomain)

def postmsg(message):
    """Sends the message to the jlogfile logging stream at level INFO.
    This is identical to:
      jlogger.info(message)."""
    return jlogger.info(message)

class MasterLogFormatter(logging.Formatter): 
    """This is a custom log formatter that inserts the thread or
    process (logthread) that generated the log message.  Also, it
    always directly calls formatException from format, ensuring that
    cached information is not used.  That allows a subclass
    (JLogFormatter) to ignore exceptions.""" 
    def __init__(self,fmt=None,datefmt=None,logthread=None): 
        logging.Formatter.__init__(self,fmt=fmt,datefmt=datefmt)
        self._logthread=None 
    @property
    def logthread(self):
        """The name of the batch thread or process that generated log
        messages, if the LogRecord does not supply that already.""" 
        global logthread # use global value if I don't have one 
        if self._logthread is None: return logthread 
        return self._logthread 
    def format(self, record):
        # We need to override this due to a "feature" in the
        # Formatter.format: It ignores formatException (never calls
        # it) and caches the exception info, even if the formatter is
        # not supposed to output it.
        global logthread
        record.message = record.getMessage()
        if self._fmt.find("%(asctime)") >= 0:
            record.asctime = self.formatTime(record, self.datefmt)
        if 'logthread' not in record.__dict__:
            record.__dict__['logthread']=self.logthread
        s = self._fmt % record.__dict__
        if 'exc_info' in record.__dict__ and record.exc_info is not None:
            e = self.formatException(record.exc_info)
            if e:
                rec2=dict(record.__dict__)
                for line in str(e).splitlines():
                    rec2['message']=line
                    s="%s\n%s"%( s, self._fmt % rec2 )
        return s

class JLogFormatter(MasterLogFormatter):
    """This subclass of MasterLogFormatter does not include exception
    information in the log file.  This is done to prevent cluttering
    of the log file."""
    def formatException(self, ei):
        """Returns nothing to indicate no exception information should
        be printed."""

def stdout_is_stderr():
    """Returns True if it can determine that stdout and stderr are the
    same file or terminal.  Returns False if it can determine they are
    not, or if the result is inconclusive."""
    try:
        return os.fstat(sys.stdout.fileno()) == os.fstat(sys.stderr.fileno())
    except Exception as e:
        pass
    return False

class MasterLogHandler(logging.Handler):
    """This is a custom logging Handler class used for multi-process or
multi-job batch scripts.  It has a higher minimum log level for
messages not sent to the jlogfile domain.  Also, for every log
message, the log file is opened, the message is written and the file
is closed.  This is done to mimic the postmsg command.  Exception
information is never sent to the log file."""
    def __init__(self,logger,jlogdomain,otherlevels,joformat,jformat):
        logging.Handler.__init__(self)
        self._logger=logger
        self._otherlevels=otherlevels
        self._jlogdomain=jlogdomain
        self._joformat=joformat
        self._jformat=jformat
    def stringify_record(self,record):
        assert(isinstance(self._joformat,MasterLogFormatter))
        assert(isinstance(self._jformat,MasterLogFormatter))
        global logthread
        if record.name==self._jlogdomain:
            # Reformat for jlogdomain:
            message=self._jformat.format(record)
        elif record.levelno<self._otherlevels:
            return None # log level too low
        else:
            message=self._joformat.format(record)
        message+='\n'
        return message
    def emit(self,record):
        message=self.stringify_record(record)
        self._logger.write(message)

class JLogHandler(MasterLogHandler):
    """This is a custom loging Handler class for the jlogfile.  It has a
higher minimum log level for messages not sent to the jlogfile domain.
Also, for every log message, the log file is opened, the message is written
and the file is closed.  This is done to mimic the postmsg command.  
Exception information is never sent to the log file."""
    def emit(self,record):
        message=self.stringify_record(record)
        if message is None: return
        if isinstance(self._logger,basestring):
            # Open the file, write and close it, to mimic the postmsg script:
            with open(self._logger,'at') as f:
                f.write(message)
        else:
            self._logger.write(message)

def mpi_redirect(threadname,stderrfile,stdoutfile,
                 masterlevel=logging.WARNING,
                 openmode=None,logger=None):
    """When the Python script splits itself into multiple processes
    via MPI, this function is called to redirect stdout to stdoutfile,
    stderr to stderrfile, and produce a new logging stream to the
    original stderr, with a logging level set to masterlevel.  That
    new logging stream is called the "master log" and will receive any
    messages at level masterlevel or higher, and any messages sent to
    the jlogdomain.

    This can also be used to redirect ONLY stdout, in which case no
    master logging stream is set up.  That is requested by
    stderrfile=None."""
    if logger is None: logger=logging.getLogger('produtil')
    if openmode is None:
        openmode=os.O_CREAT|os.O_WRONLY|os.O_APPEND
    elif not isinstance(openmode,int):
        raise TypeError(
            "In mpiRedirect, the openmode must be an int, not a "
            +type(openmode).__name__)
    if not isinstance(stdoutfile,basestring):
        raise TypeError(
            "In mpiRedirect, the stdoutfile must be a string, not a "
            +type(stdoutfile).__name__)
    if stderrfile is not None and not isinstance(stderrfile,basestring):
        raise TypeError(
            "In mpiRedirect, the stderrfile must be a string or None, not a "
            +type(stderrfile).__name__)
    if not isinstance(threadname,basestring):
        raise TypeError(
            "In mpiRedirect, the threadname must be a string, not a "
            +type(threadname))

    global logthread
    logthread='['+str(threadname)+']'

    logger.warning('Redirecting stdout to "%s" for thread %s'
                   %(stdoutfile,logthread))
    fd=os.open(stdoutfile,openmode)
    os.dup2(fd,1)

    if stderrfile is not None:
        logger.warning('Redirecting stderr to "%s" for thread %s'
                       %(stderrfile,logthread))

        olderrfd=os.dup(2)
        olderr=os.fdopen(olderrfd,'at',0)

        if(stdoutfile!=stderrfile):
            # Only reopen if stderr differs from stdout.
            fd=os.open(str(stderrfile),openmode)
        os.dup2(fd,2)
    
        oformat=MasterLogFormatter(
            "%(asctime)s.%(msecs)03d %(name)s %(logthread)s (%(filename)s:"
            "%(lineno)d) %(levelname)s: %(message)s",
            "%m/%d %H:%M:%S")

        # Create the master log handler.  It will send to the old stderr
        # (from before the redirection), and send everything from
        # masterlevel and higher from any logging domain.  Anything sent
        # to the jlogfile, at any level, will be sent to the master log
        # file as well.
        mloghandler=MasterLogHandler(olderr,jlogdomain,masterlevel,oformat,
                                     oformat)
        logger.warning('Turning on logging of high priority messages to '
                       'original stderr stream.')
        if masterlevel!=logging.NOTSET: mloghandler.setLevel(masterlevel)
        logging.getLogger().addHandler(mloghandler)

def configureLogging(jlogfile=None,
                     level=logging.INFO, # all messages filtered by this
                     jloglevel=logging.INFO,
                     japplevel=logging.ERROR, # should be >=jloglevel
                     eloglevel=logging.WARNING,
                     ologlevel=logging.NOTSET):
    """Configures log output to stderr, stdout and the jlogfile
    Options:
        jlogfile = path to the jlogfile. Default: use
            os.environ('jlogfile') if set.  Otherwise, stderr.
        level = minimum logging level globally.  Set to INFO by default.
            Change this to logging.DEBUG if you're debugging the program.
        jloglevel = minimum logging level to send to jlogfile
        japplevel = minimum logging level to send to jlogfile from all
            domains except that specified in jlogdomain.  Be careful
            when changing this as it logs directly to the WCOSS-wide
            jlogfile in operations.
        eloglevel = minimum logging level to send to stderr from ALL logs
            Set to None to disable stderr logging
        ologlevel = minimum logging level to send to stdout from ALL logs
            Default: logging.NOTSET (no filtering)
            Set to None to disable stdout logging.
    Important notes when choosing levels:
      level - sets the global minimum log level.  Anything below this
          level will be discarded regardless of other settings.
      jloglevel - this limit is applied before japplevel"""
    root=logging.getLogger()
    if level!=logging.NOTSET:
        root.setLevel(level) # set global minimum logging level

    # Configure log formatting:
    jlog=logging.getLogger('jlogfile')
    jobstr=os.environ.get('job',None)
    if jobstr is None:
        jobstr=produtil.batchsystem.jobname()
    jobstr=str(jobstr).replace('(','_').replace(')','_').replace('%','_')
    # Format for jlogfile domain logging to jlogfile:
    jformat=JLogFormatter(
        "%(asctime)sZ "+jobstr+"-%(levelname)s: %(logthread)s%(message)s",
        "%m/%d %H:%M:%S")
    # Format for other domains logging to jlogfile is the same, but
    # with the domain added:
    joformat=JLogFormatter(
        "%(asctime)sZ "+jobstr+
        "-%(name)s: %(levelname)s: %(logthread)s%(message)s",
        "%m/%d %H:%M:%S")
    # For stdout/stderr, a more verbose version with milliseconds,
    # file and line numbers:
    oformat=logging.Formatter(
        "%(asctime)s.%(msecs)03d %(name)s (%(filename)s:%(lineno)d) "
        "%(levelname)s: %(message)s",
        "%m/%d %H:%M:%S")

    if stdout_is_stderr():
        # Configure combined stdout+stderr logging:
        loglevel=min(ologlevel,eloglevel)
        logstream=logging.StreamHandler(sys.stderr)
        logstream.setFormatter(oformat)
        if loglevel!=logging.NOTSET: logstream.setLevel(loglevel)
        root.addHandler(logstream)
    else:
        # Configure stdout logging:
        if ologlevel is not None:
            ologstream=logging.StreamHandler(sys.stdout)
            ologstream.setFormatter(oformat)
            if ologlevel!=logging.NOTSET: ologstream.setLevel(ologlevel)
            root.addHandler(ologstream)

        # Configure stderr logging:
        if eloglevel is not None:
            elogstream=logging.StreamHandler(sys.stderr)
            elogstream.setFormatter(oformat) # same format as stdout
            if eloglevel!=logging.NOTSET: elogstream.setLevel(eloglevel)
            root.addHandler(elogstream)

    # Configure jlogfile logging:
    #   jlogfile domain: INFO and higher
    #   all domains: ERROR and higher
    if jlogfile is None:
        # Try to get the jlogfile from the environment if none is specified:
        var=str(os.environ.get('jlogfile',''))
        if len(var)>0: jlogfile=var
    # If we still don't have the jlogfile, use stderr:
    jlogfile=str(jlogfile) if jlogfile is not None else sys.stderr
    jloghandler=JLogHandler(jlogfile,jlogdomain,japplevel,joformat,jformat)
    if jloglevel!=logging.NOTSET: jloghandler.setLevel(jloglevel)

    root.addHandler(jloghandler)
