"""This module is a workaround for a defficiency of Python.  When
Python receives a fatal signal other than SIGINT, it exits immediately
without freeing utilized resources or otherwise cleaning up.  This
module causes Python to raise a fatal exception, that does NOT derive
from Exception, if a fatal signal is received.  Note that there is a
critical flaw in this design: raising an exception in a signal handler
only raises it in the main (initial) thread.  Other threads must call
the produtil.sigsafety.checksig function as frequently as possible to
check if a signal has been caught.  That function will raise the
appropriate exception if a signal was caught, or return immediately
otherwise.

The reason this HAD to be added to produtil is that the lack of proper
signal handling caused major problems.  In particular, it completely
broke file locking on Lustre and Panasas.  Both filesystems will
sometimes forget a file lock is released if the lock was held by a
process that exited abnormally.  There were also unverified cases of
this happening with GPFS.  Correctly handling SIGTERM, SIGQUIT, SIGHUP
and SIGINT has solved that problem thus far.

The base class of any exeption thrown due to a signal is CaughtSignal.
It has two subclasses: FatalSignal, which is raised when a fatal
signal is received, and HangupSignal.  The HangupSignal is raised by
SIGHUP, unless the install_handlers requests otherwise.  Scripts
should catch HangupSignal if the program is intended to ignore
hangups.  However, nothing shoudl ever catch FatalSignal.  Only
__exit__ and finalize blocks should be run in that situation, and they
should run as quickly as possible.

The install_handlers installs the signal handlers: term_handler and
optionally hup_handler.  The raise_signals option specifies the list
of signals that will raise FatalSignal, defaulting to SIGTERM, SIGINT
and SIGQUIT.  If SIGHUP is added to that list, then it will raise
FatalSignal as well.  Otherwise, the ignore_hup option controls the
SIGHUP behavior: if True, SIGHUP is simply ignored, otherwise it
raises HangupSignal.

One can call install_handlers directly, though it is recommended to
call produtil.setup.setup instead."""

import produtil.locking
import signal

__all__=['CaughtSignal','HangupSignal','FatalSignal','install_handlers','checksig']

class CaughtSignal(KeyboardInterrupt):
    """Base class of the exceptions thrown when a signal is caught.
    Note that this does not derive from Exception, to ensure it is not
    caught accidentally.  At present, it derives directly from
    KeyboardInterrupt, though that may be changed in the future to
    BaseException."""
    def __init__(self,signum):
        BaseException.__init__(self)
        self.signum=signum
    def __str__(self):
        return 'Caught signal %d'%(self.signum,)
class HangupSignal(CaughtSignal):
    """With the default settings to install_handlers, this is raised
    when a SIGHUP is caught.  Note that this does not derive from
    Exception."""
class FatalSignal(CaughtSignal):
    """Raised when a fatal signal is caught, as defined by the call to
    install_handlers.  Note that this does not derive from
    Exception."""

caught_signal=None
"""The signal number of the signal that was caught or None if no
signal has been caught.  This is initialized by the signal handlers,
and used by checksig to raise exceptions due to caught signals."""

caught_class=None
"""The class that should be raised due to the caught signal, or None
if no signal has been caught.  This is initialized by the signal
handlers, and used by checksig to raise exceptions due to caught
signals."""

def checksig():
    """This should be called frequently from worker threads to
    determine if the main thread has received a signal.  If a signal
    was caught this function will raise the appropriate subclass of
    CaughtSignal.  Otherwise, it returns None."""
    global caught_signal,caught_class
    cs=caught_signal
    cc=caught_class
    if cs is not None and cc is not None:
        raise cc(cs)
    return None

def hup_handler(signum,frame):
    """This is the signal handler for raising HangupSignal: it is used
    only for SIGHUP, and only if that is not specified in
    raise_signals and ignore_hup=False."""
    global caught_signal,caught_class
    caught_signal=signum
    caught_class=HangupSignal

    produtil.locking.disable_locking()
    raise HangupSignal(signum)

def term_handler(signum,frame): 
    """This is the signal handler for raising FatalSignal."""
    global caught_signal,caught_class
    caught_signal=signum
    caught_class=HangupSignal

    produtil.locking.disable_locking()
    raise FatalSignal(signum)

def install_handlers(ignore_hup=False,raise_signals=[signal.SIGTERM,signal.SIGINT,signal.SIGQUIT]):
    """Installs signal handlers that will raise exceptions.

        ignore_hup - If True, SIGHUP is ignored, else SIGHUP will
            raise HangupSignal

        raise_signals - List of exceptions that will raise
            FatalSignal.  If SIGHUP is in this list, that overrides
            any decision made through ignore_hup.
    """
    if(ignore_hup):
        signal.signal(signal.SIGHUP,signal.SIG_IGN)
    elif signal.SIGHUP not in raise_signals:
        signal.signal(signal.SIGHUP,hup_handler)
    for sig in raise_signals:
        signal.signal(sig,term_handler)
